"""

Logger that writes xlsx files

"""

import datetime
import logging
import numpy
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import logging
import sys
import yaml
import copy
import gzip
import os
import netCDF4
import pydantic
import typing
from redvypr.device import RedvyprDevice
import redvypr.data_packets as data_packets
import redvypr.redvypr_address as redvypr_address
import redvypr.packet_statistic as packet_statistics
import redvypr.gui

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('netcdfwriter')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True
class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = "Saves subscribed devices in a netCDF4 file"
    gui_tablabel_display: str = 'netCDF logging status'

class DeviceCustomConfig(pydantic.BaseModel):
    dt_sync: int = pydantic.Field(default=5,description='Time after which an open file is synced on disk')
    dt_newfile: int = pydantic.Field(default=3600,description='Time after which a new file is created')
    dt_newfile_unit: typing.Literal['none','seconds','hours','days'] = pydantic.Field(default='seconds')
    dt_update:int = pydantic.Field(default=2,description='Time after which an upate is sent to the gui')
    clearqueue: bool = pydantic.Field(default=True, description='Flag if the buffer of the subscribed queue should be emptied before start')
    zlib: bool = pydantic.Field(default=True, description='Flag if zlib compression shall be used for the netCDF data')
    size_newfile:int = pydantic.Field(default=500,description='Size of object in RAM after which a new file is created')
    size_newfile_unit: typing.Literal['none','bytes','kB','MB'] = pydantic.Field(default='MB')
    datafolder:str = pydantic.Field(default='./',description='Folder the data is saved to')
    fileextension:str= pydantic.Field(default='nc',description='File extension, if empty not used')
    fileprefix:str= pydantic.Field(default='redvypr',description='If empty not used')
    filepostfix:str= pydantic.Field(default='netcdflogger',description='If empty not used')
    filedateformat:str= pydantic.Field(default='%Y-%m-%d_%H%M%S',description='Dateformat used in the filename, must be understood by datetime.strftime')
    filecountformat:str= pydantic.Field(default='04',description='Format of the counter. Add zero if trailing zeros are wished, followed by number of digits. 04 becomes {:04d}')

def create_logfile(config,count=0):
    funcname = __name__ + '.create_logfile():'
    logger.debug(funcname)

    filename = ''
    if len(config['datafolder']) > 0:
        if os.path.isdir(config['datafolder']):
            filename += config['datafolder'] + os.sep
        else:
            logger.warning(funcname + ' Data folder {:s} does not exist.'.format(filename))
            return None

    if(len(config['fileprefix'])>0):
        filename += config['fileprefix']

    if (len(config['filedateformat']) > 0):
        tstr = datetime.datetime.now().strftime(config['filedateformat'])
        filename += '_' + tstr

    if (len(config['filecountformat']) > 0):
        cstr = "{:" + config['filecountformat'] +"d}"
        filename += '_' + cstr.format(count)

    if (len(config['filepostfix']) > 0):
        filename += '_' + config['filepostfix']

    if (len(config['fileextension']) > 0):
        filename += '.' + config['fileextension']

    logger.info(funcname + ' Will create a new file: {:s}'.format(filename))

    # Create a workbook and add a worksheet.
    nc = netCDF4.Dataset(filename, mode='w',format='NETCDF4')
    return [nc,filename]

def start(device_info, config, dataqueue=None, datainqueue=None, statusqueue=None):
    logger_start = logging.getLogger('netcdfwriter/thread')
    logger_start.setLevel(logging.DEBUG)
    funcname = __name__ + '.start()'
    logger_start.debug(funcname + ':Opening writing:')
    #print('Config',config)
    if config['clearqueue']:
        while (datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
            except:
                break

    data_write_to_file = [] # List of columns to be written to file
    count = 0
    if True:
        try:
            dtneworig  = config['dt_newfile']
            dtunit     = config['dt_newfile_unit']
            if(dtunit.lower() == 'seconds'):
                dtfac = 1.0
            elif(dtunit.lower() == 'hours'):
                dtfac = 3600.0
            elif(dtunit.lower() == 'days'):
                dtfac = 86400.0
            else:
                dtfac = 0
                
            dtnews     = dtneworig * dtfac
            logger_start.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['dt_newfile'],config['dt_newfile_unit']))
        except Exception as e:
            logger.exception(e)
            dtnews = 0
            
        try:
            sizeneworig  = config['size_newfile']
            sizeunit = config['size_newfile_unit']
            if(sizeunit.lower() == 'kb'):
                sizefac = 1000.0
            elif(sizeunit.lower() == 'mb'):            
                sizefac = 1e6
            elif(sizeunit.lower() == 'bytes'):            
                sizefac = 1 
            else:
                sizefac = 0
                
            sizenewb     = sizeneworig * sizefac # Size in bytes
            logger_start.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['size_newfile'],config['size_newfile_unit']))
        except Exception as e:
            logger.exception(e)
            sizenewb = 0  # Size in bytes
            
            
    try:
        config['dt_sync']
    except:
        config['dt_sync'] = 5


    flag_zlib = config['zlib']
    #flag_zlib = False
    bytes_written = 0
    packets_written = 0
    bytes_written_total = 0
    packets_written_total = 0
    [nc,filename] = create_logfile(config,count)
    logger_start.debug('Adding main group {}'.format(device_info))
    hostname = device_info['hostinfo']['hostname']
    redvypr_version_str = 'redvypr {}'.format(redvypr.version)
    nc.redvypr_version = redvypr_version_str
    data_stat = {'_deviceinfo': {}}
    data_stat['_deviceinfo']['filename'] = filename
    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
    data_stat['_deviceinfo']['created'] = time.time()
    data_stat['_deviceinfo']['bytes_written'] = bytes_written
    data_stat['_deviceinfo']['packets_written'] = packets_written
    dataqueue.put(data_stat)
    count += 1

    tfile = time.time() # Save the time the file was created
    tflush = time.time() # Save the time the file was created
    tupdate = time.time() # Save the time for the update timing
    FLAG_RUN = True
    file_status = {}
    file_status_reduced = file_status
    deviceinfo_all = None
    metadata_update = False
    while FLAG_RUN:
        tcheck      = time.time()
        # Flush file on regular basis
        if ((time.time() - tflush) > config['dt_sync']):
            logger_start.info(funcname + 'Syncing netCDF file {}'.format(filename))
            nc.sync()
            bytes_written = os.path.getsize(filename)
            #bytes_written = pympler.asizeof.asizeof(nc)
            #print('Bytes written',bytes_written)
            tflush = time.time()

        time.sleep(0.05)
        while(datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
                packet_address = redvypr.RedvyprAddress(data)
                if (data is not None):
                    [command,comdata] = data_packets.check_for_command(data, thread_uuid=device_info['thread_uuid'], add_data=True)
                    #logger.debug('Got a command: {:s}'.format(str(data)))
                    if (command is not None):
                        if(command == 'stop'):
                            logger_start.debug('Stop command')
                            FLAG_RUN = False
                            break

                        if (command == 'info'):
                            logger_start.debug('Metadata command')
                            if packet_address.packetid == 'metadata':
                                deviceinfo_all = data['deviceinfo_all']
                                metadata_update = True
                                vars_updated = []

                    # Ignore some packages
                    if data in redvypr_address.RedvyprAddress(redvypr.metadata_address):
                        logger_start.debug('Ignoring metadata packet')
                        continue


                #statistics = data_packets.do_data_statistics(data,statistics)
                #packet_address = redvypr.RedvyprAddress(data)
                # Check if host uuid is the same as local uuid
                if packet_address.uuid == device_info['hostinfo']['uuid']:
                    print('Local')
                    print('Local')
                    print('Local')
                else:
                    print('Packet data', data)
                    print('Packet address', packet_address.get_fullstr())
                    print('Packet address hostname',packet_address.hostname)
                    hostname = packet_address.hostname + '__UUID__' + packet_address.uuid
                    print('Hostname',hostname)
                    print('Remote')
                    print('Remote')
                    print('Remote')
                address_format = '/h/p/d/'
                packet_address_str = packet_address.get_str(address_format)
                publisher = packet_address.publisher
                devicename = packet_address.devicename
                # This is the group structure
                # Data is written to the group found in
                # ncgroup = groups[hostname][publisher][devicename]

                try:
                    nc[hostname]
                except:
                    logger_start.debug('Creating base group {}'.format(hostname))
                    nchost = nc.createGroup(hostname)

                try:
                    nc[hostname][publisher]
                except:
                    logger_start.debug('Creating publishing device {}'.format(publisher))
                    ncgroup_pub = nc[hostname].createGroup(publisher)

                # The device
                try:
                    nc_device = nc[hostname][publisher][devicename]
                except:
                    logger_start.debug('Creating device {}'.format(devicename))
                    nc_device = nc[hostname][publisher].createGroup(devicename)
                    nc_device.redvypr_address = redvypr_address.RedvyprAddress(data).get_fullstr()
                    # Add time variable
                    logger.debug('Creating time dimension')
                    nc_device.createDimension('time',None)
                    nc_device.createVariable('time', float,('time'))

                # Write metadata of roogroup and devices
                if deviceinfo_all is not None and not(nc in vars_updated):
                    # Write metadata
                    try:
                        if deviceinfo_all is not None and not (var in vars_updated):
                            raddress_tmp = redvypr_address.RedvyprAddress(data)
                            raddress_tmp_str = raddress_tmp.get_str('/h/d/i')
                            raddress_tmp_str_full = raddress_tmp.get_fullstr()
                            metadata_tmp = packet_statistics.get_metadata_deviceinfo_all(deviceinfo_all, raddress_tmp)
                            # print('Metadata tmp', raddress_tmp, metadata_tmp)
                            # device_worksheets[packet_address_str].write(lineindex, colindex, datawrite)
                            if len(metadata_tmp.keys()) > 0:  # Check if something was found
                                for metakey in metadata_tmp.keys():
                                    setattr(nc_device, metakey, metadata_tmp[metakey])

                            vars_updated.append(var)
                    except:
                        logger_start.debug('Could not set metadata', exc_info=True)


                # Write data
                if True:
                    #print(funcname + ' got data',data)
                    try:
                        data['t']
                    except:
                        data['t'] = data['_redvypr']['t']

                    lent = len(nc_device.variables['time'])
                    nc_device.variables['time'][lent] = data['t']
                    datakeys = data_packets.Datapacket(data).datakeys()
                    datakeys.remove('t')
                    #datakeys.insert(0,'t')
                    # Write data in datakeys or create variable
                    # Write time
                    packets_written += 1
                    for k in datakeys:
                        print('-----')
                        print('Datakeys', datakeys)
                        print('Datakey',k)
                        try:
                            var = nc_device.variables[k]
                        except: # Create variable
                            typedata = type(data[k])
                            #print('typedata',typedata)
                            if (typedata is list) or (typedata is numpy.ndarray):
                                try:
                                    logger_start.info('Creating variable for list/ndarray type {}'.format(typedata))
                                    dwrite = numpy.asarray(data[k])
                                    datatype_array = dwrite.dtype
                                    dwrite_shape = numpy.shape(dwrite)
                                    dimnames = ['time']
                                    for id,nd in enumerate(numpy.shape(dwrite)):
                                        dimname = k + '_n_{}'.format(id)
                                        dimnames.append(dimname)
                                        nc_device.createDimension(dimname, None)

                                    logger_start.debug('Creating variable {}. Dimnames {}. Datatype {}.'.format(k,dimnames,datatype_array))
                                    var = nc_device.createVariable(k, datatype_array, dimnames, zlib=flag_zlib)
                                    setattr(var, 'redvypr_address', packet_address.get_fullstr())
                                except:
                                    logger_start.warning('Could not create variable for {}'.format(k),exc_info=True)
                            elif (typedata is str):
                                logger_start.info('Creating string variable')
                                # For some reason zlib does not work with str
                                var = nc_device.createVariable(k, str, ('time'), zlib=False)
                                setattr(var, 'redvypr_address', packet_address.get_fullstr())
                            else:
                                try:
                                    logger_start.info('Creating variable with type {}'.format(typedata))
                                    var = nc_device.createVariable(k, typedata, ('time'), zlib=flag_zlib)
                                    setattr(var, 'redvypr_address', packet_address.get_fullstr())
                                except:
                                    var = None

                        if var is not None:
                            try:
                                var[lent] = data[k]
                                try:
                                    file_status[k] += 1
                                except:
                                    file_status[k] = 1
                            except:
                                logger_start.warning('Could not write data',exc_info=True)
                        else:
                            pass


                        # Write metadata
                        try:
                            if deviceinfo_all is not None and not(var in vars_updated):
                                raddress_tmp = redvypr_address.RedvyprAddress(data, datakey=k)
                                raddress_tmp_str = raddress_tmp.get_str('/h/d/i/k')
                                raddress_tmp_str_full = raddress_tmp.get_fullstr()
                                metadata_tmp = packet_statistics.get_metadata_deviceinfo_all(deviceinfo_all, raddress_tmp)
                                # print('Metadata tmp', raddress_tmp, metadata_tmp)
                                # device_worksheets[packet_address_str].write(lineindex, colindex, datawrite)
                                if len(metadata_tmp.keys()) > 0:  # Check if something was found
                                    for metakey in metadata_tmp.keys():
                                        setattr(var,metakey,metadata_tmp[metakey])

                                vars_updated.append(var)
                        except:
                            logger_start.debug('Could not set metadata',exc_info=True)


                # Send statistics
                if ((time.time() - tupdate) > config['dt_update']):
                    tupdate = time.time()
                    data_stat = {'_deviceinfo': {}}
                    data_stat['_deviceinfo']['filename'] = filename
                    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                    data_stat['_deviceinfo']['bytes_written'] = bytes_written
                    data_stat['_deviceinfo']['packets_written'] = packets_written
                    data_stat['_deviceinfo']['file_status'] = file_status
                    data_stat['_deviceinfo']['file_status_reduced'] = file_status_reduced
                    dataqueue.put(data_stat)

            except Exception as e:
                logger.exception(e)
                logger.debug(funcname + ':Exception:' + str(e))
                # print(data)

        if True: # Check if a new file should be created, close the old one and write the header
            file_age = tcheck - tfile
            FLAG_TIME = (dtnews > 0) and (file_age >= dtnews)
            FLAG_SIZE = (sizenewb > 0) and (bytes_written >= sizenewb)
            if FLAG_TIME or FLAG_SIZE or (FLAG_RUN == False):
                # Autofit
                nc.close()
                data_stat = {'_deviceinfo': {}}
                data_stat['_deviceinfo']['filename'] = filename
                data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                data_stat['_deviceinfo']['closed'] = time.time()
                data_stat['_deviceinfo']['bytes_written'] = bytes_written
                data_stat['_deviceinfo']['packets_written'] = packets_written
                dataqueue.put(data_stat)
                if FLAG_RUN:
                    tfile = tcheck
                    bytes_written = 0
                    packets_written = 0
                    bytes_written_total = 0
                    packets_written_total = 0
                    [nc, filename] = create_logfile(config, count)
                    redvypr_version_str = 'redvypr {}'.format(redvypr.version)

                    data_stat = {'_deviceinfo': {}}
                    data_stat['_deviceinfo']['filename'] = filename
                    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                    data_stat['_deviceinfo']['created'] = time.time()
                    data_stat['_deviceinfo']['bytes_written'] = bytes_written
                    data_stat['_deviceinfo']['packets_written'] = packets_written
                    dataqueue.put(data_stat)

class Device(RedvyprDevice):
    """
    netCDFlogger device
    """

    def __init__(self, **kwargs):
        """
        """
        funcname = __name__ + '__init__()'
        super(Device, self).__init__(**kwargs)


#
#
# The init widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QGridLayout(self)
        #print('Hallo,device config',device.config)
        self.device   = device
        self.redvypr  = device.redvypr
        self.label    = QtWidgets.QLabel("netCDF logger setup")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet(''' font-size: 24px; font: bold''')
        self.config_widgets= [] # A list of all widgets that can only be used of the device is not started yet
        # Input output widget
        self.inlabel  = QtWidgets.QLabel("Input")
        self.inlist   = QtWidgets.QListWidget()
        #
        self.adddeviceinbtn   = QtWidgets.QPushButton("Subscribe")
        self.adddeviceinbtn.clicked.connect(self.con_clicked)
        self.adddeviceinbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        # The output widgets
        self.outlabel        = QtWidgets.QLabel("Logfile")
        self.outfilename     = QtWidgets.QLineEdit()
        # Checkboxes
        self.prefix_check    = QtWidgets.QCheckBox('Prefix')
        self.date_check      = QtWidgets.QCheckBox('Date/Time')
        self.count_check     = QtWidgets.QCheckBox('Counter')
        self.postfix_check   = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')

        try:
            filename = self.device.custom_config['filename']
        except:
            filename = ''

        self.outfilename.setText(filename)
        
        self.folderbtn   = QtWidgets.QPushButton("Folder")
        self.config_widgets.append(self.folderbtn)
        self.folderbtn.clicked.connect(self.get_datafolder)
        
        # The rest
        #self.conbtn = QtWidgets.QPushButton("Connect logger to devices")
        #self.conbtn.clicked.connect(self.con_clicked)        
        self.startbtn = QtWidgets.QPushButton("Start logging")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred,QtWidgets.QSizePolicy.Expanding)


        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.dt_newfile = edit
        self.dt_newfile.setToolTip('Create a new file every N seconds.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.dt_newfile.setText(str(self.device.custom_config['dt_newfile']))
        except Exception as e:
            self.dt_newfile.setText('0')
            
        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.size_newfile = edit
        self.size_newfile.setToolTip('Create a new file if N bytes of RAM are used.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.size_newfile.setText(str(self.device.custom_config.size_newfile))
        except Exception as e:
            self.size_newfile.setText('0')
            
        self.newfiletimecombo = QtWidgets.QComboBox()
        times = typing.get_args(self.device.custom_config.model_fields['dt_newfile_unit'].annotation)
        timeunit = self.device.custom_config.dt_newfile_unit
        for t in times:
            self.newfiletimecombo.addItem(t)

        index = self.newfiletimecombo.findText(timeunit)
        self.newfiletimecombo.setCurrentIndex(index)
        #self.newfiletimecombo.addItem('None')
        #self.newfiletimecombo.addItem('seconds')
        #self.newfiletimecombo.addItem('hours')
        #self.newfiletimecombo.addItem('days')
        #self.newfiletimecombo.setCurrentIndex(1)
            
        self.newfilesizecombo = QtWidgets.QComboBox()
        self.newfilesizecombo.addItem('None')
        self.newfilesizecombo.addItem('bytes')
        self.newfilesizecombo.addItem('kB')
        self.newfilesizecombo.addItem('MB')
        self.newfilesizecombo.setCurrentIndex(3)
        
        sizelabel = QtWidgets.QLabel('New file after')
         # File change layout
        self.newfilewidget = QtWidgets.QWidget()
        self.newfilelayout = QtWidgets.QFormLayout(self.newfilewidget)
        self.newfilelayout.addRow(sizelabel)
        self.newfilelayout.addRow(self.dt_newfile,self.newfiletimecombo)
        self.newfilelayout.addRow(self.size_newfile,self.newfilesizecombo)
        
        # Filenamelayout
        self.folder_text = QtWidgets.QLineEdit('')
        self.extension_text = QtWidgets.QLineEdit('redvypr_raw')
        self.prefix_text = QtWidgets.QLineEdit('')
        self.date_text = QtWidgets.QLineEdit('%Y-%m-%d_%H%M%S')
        self.count_text = QtWidgets.QLineEdit('04d')
        self.postfix_text = QtWidgets.QLineEdit('')

        self.prefix_check = QtWidgets.QCheckBox('Prefix')
        self.date_check = QtWidgets.QCheckBox('Date/Time')
        self.count_check = QtWidgets.QCheckBox('Counter')
        self.postfix_check = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')
        # The outwidget
        self.outwidget = QtWidgets.QWidget()
        self.outlayout = QtWidgets.QGridLayout(self.outwidget)
        # Datafolder lineedit
        self.outlayout.addWidget(self.folderbtn, 0, 0)
        self.outlayout.addWidget(self.folder_text, 0, 1,1,4)
        # Checkboxes
        self.outlayout.addWidget(self.prefix_check, 1, 0)
        self.outlayout.addWidget(self.date_check, 1, 1)
        self.outlayout.addWidget(self.count_check, 1, 2)
        self.outlayout.addWidget(self.postfix_check, 1, 3)
        self.outlayout.addWidget(self.extension_check, 1, 4)

        self.outlayout.addWidget(self.prefix_text, 2, 0)
        self.outlayout.addWidget(self.date_text, 2, 1)
        self.outlayout.addWidget(self.count_text, 2, 2)
        self.outlayout.addWidget(self.postfix_text, 2, 3)
        self.outlayout.addWidget(self.extension_text, 2, 4)

        self.outlayout.addWidget(self.newfilewidget,4,0,1,4)

        #self.outlayout.addStretch(1)
            
        layout.addWidget(self.label,0,0,1,2)
        #layout.addWidget(self.inlabel,1,0)
        #layout.addWidget(self.inlist,2,0)
        layout.addWidget(self.outlabel,1,0)
        layout.addWidget(self.outwidget,2,0)
        layout.addWidget(self.adddeviceinbtn, 5, 0)
        layout.addWidget(self.startbtn,6,0)

        self.config_to_widgets()
        self.connect_widget_signals()
        # Connect the signals that notify a change of the connection
        self.device.redvypr.device_status_changed_signal.connect(self.update_device_list)
        #self.redvypr.devices_connected.connect
        self.device.subscription_changed_signal.connect(self.update_device_list)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)


    def connect_widget_signals(self,connect=True):
        """
        Connects the signals of the widgets such that an update of the config is done

        Args:
            connect:

        Returns:

        """
        funcname = self.__class__.__name__ + '.connect_widget_signals():'
        logger.debug(funcname)
        if(connect):
            self.prefix_check.stateChanged.connect(self.update_device_config)
            self.postfix_check.stateChanged.connect(self.update_device_config)
            self.date_check.stateChanged.connect(self.update_device_config)
            self.count_check.stateChanged.connect(self.update_device_config)
            self.extension_check.stateChanged.connect(self.update_device_config)
            self.prefix_text.editingFinished.connect(self.update_device_config)
            self.postfix_text.editingFinished.connect(self.update_device_config)
            self.date_text.editingFinished.connect(self.update_device_config)
            self.count_text.editingFinished.connect(self.update_device_config)
            self.extension_text.editingFinished.connect(self.update_device_config)
            self.newfilesizecombo.currentIndexChanged.connect(self.update_device_config)
            self.newfiletimecombo.currentIndexChanged.connect(self.update_device_config)
        else:
            self.prefix_check.stateChanged.disconnect()
            self.postfix_check.stateChanged.disconnect()
            self.date_check.stateChanged.disconnect()
            self.count_check.stateChanged.disconnect()
            self.extension_check.stateChanged.disconnect()
            self.prefix_text.editingFinished.disconnect()
            self.postfix_text.editingFinished.disconnect()
            self.date_text.editingFinished.disconnect()
            self.count_text.editingFinished.disconnect()
            self.extension_text.editingFinished.disconnect()
            self.newfilesizecombo.currentIndexChanged.disconnect()
            self.newfiletimecombo.currentIndexChanged.disconnect()


    def get_datafolder(self):
        funcname = self.__class__.__name__ + '.get_datafolder():'
        logger.debug(funcname)
        retdata = QtWidgets.QFileDialog.getExistingDirectory(self,"Choose datafolder")
        #print('Datafolder',retdata)
        datafolder = retdata
        if datafolder:
            self.folder_text.setText(datafolder)
            
    def con_clicked(self):
        funcname = self.__class__.__name__ + '.con_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if(button == self.adddeviceinbtn):
            self.connect.emit(self.device) # The connect signal is connected with connect_device that will open a subscribe/connect widget
            self.update_device_list()

    def update_datastream_table(self):
        """

        """
        # columns = ['Columnnr.', 'Datakey', 'Device', 'Format', 'Full Address String']
        funcname = self.__class__.__name__ + '.update_datastream_table():'
        print(funcname)
        datastreams_subscribed = self.device.get_subscribed_datastreams()
        print('datastreams subscribed',datastreams_subscribed)
        self.datastreamtable.clear()
        self.datastreamtable.setRowCount(len(datastreams_subscribed))
        for i,d in enumerate(datastreams_subscribed):
            dadr = redvypr_address(d)
            item = QtWidgets.QTableWidgetItem(d)
            self.datastreamtable.setItem(i,4,item)

        self.datastreamtable.resizeColumnsToContents()

    def update_device_list(self):
        funcname = self.__class__.__name__ + '.update_device_list():'
        logger.debug(funcname)
        #print('Devices',devicestr_provider,devicestr_receiver)
        raddresses = self.device.get_subscribed_deviceaddresses()
        print('Deviceaddresses',raddresses)
        self.inlist.clear()
        for raddr in raddresses:
            self.inlist.addItem(raddr.address_str)


    def config_to_widgets(self):
        """
        Updates the widgets according to the device config

        Returns:

        """
        funcname = self.__class__.__name__ + '.config_to_widgets():'
        logger.debug(funcname)

        config = self.device.custom_config
        print('config',config)
        self.dt_newfile.setText(str(config.dt_newfile))
        for i in range(self.newfiletimecombo.count()):
            self.newfiletimecombo.setCurrentIndex(i)
            if(self.newfiletimecombo.currentText().lower() == config.dt_newfile_unit):
                break

        for i in range(self.newfilesizecombo.count()):
            self.newfilesizecombo.setCurrentIndex(i)
            if (self.newfilesizecombo.currentText().lower() == config.size_newfile_unit):
                break

        self.size_newfile.setText(str(config.size_newfile))

        if len(config.datafolder)>0:
            self.folder_text.setText(config.datafolder)
        # Update filename and checkboxes
        filename_all = []
        filename_all.append([config.fileextension,self.extension_text,self.extension_check])
        filename_all.append([config.fileprefix,self.prefix_text,self.prefix_check])
        filename_all.append([config.filepostfix,self.postfix_text,self.postfix_check])
        filename_all.append([config.filedateformat,self.date_text,self.date_check])
        filename_all.append([config.filecountformat,self.count_text,self.count_check])
        for i in range(len(filename_all)):
            widgets = filename_all[i]
            if(len(widgets[0])==0):
                widgets[2].setChecked(False)
                widgets[1].setText('')
            else:
                widgets[2].setChecked(True)
                widgets[1].setText(widgets[0])

    def widgets_to_config(self,config):
        """
        Reads the widgets and creates a config
        Returns:
            config: Config dictionary
        """
        funcname = self.__class__.__name__ + '.widgets_to_config():'
        logger.debug(funcname)
        config.dt_newfile = int(self.dt_newfile.text())
        config.dt_newfile_unit = self.newfiletimecombo.currentText()
        config.size_newfile = int(self.size_newfile.text())
        config.size_newfile_unit = self.newfilesizecombo.currentText()

        config.datafolder = self.folder_text.text()

        if(self.extension_check.isChecked()):
            config.fileextension = self.extension_text.text()
        else:
            config.fileextension = ''

        if(self.prefix_check.isChecked()):
            config.fileprefix = self.prefix_text.text()
        else:
            config.fileprefix = ''

        if(self.postfix_check.isChecked()):
            config.filepostfix = self.postfix_text.text()
        else:
            config.filepostfix = ''

        if(self.date_check.isChecked()):
            config.filedateformat = self.date_text.text()
        else:
            config.filedateformat = ''

        if(self.count_check.isChecked()):
            config.filecountformat = self.count_text.text()
        else:
            config.filecountformat = ''

        print('Config',config)
        return config

    def update_device_config(self):
        """
        Updates the device config based on the widgets
        Returns:

        """
        funcname = self.__class__.__name__ + '.update_device_config():'
        logger.debug(funcname)
        self.widgets_to_config(self.device.custom_config)

    def start_clicked(self):
        funcname = self.__class__.__name__ + '.start_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if button.isChecked():
            logger.debug(funcname + "button pressed")
            self.update_device_config()
            self.device.thread_start()
        else:
            logger.debug(funcname + 'button released')
            self.device.thread_stop()

            
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']

            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                for w in self.config_widgets:
                    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                for w in self.config_widgets:
                    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        self.device = device
        layout          = QtWidgets.QVBoxLayout(self)
        hlayout         = QtWidgets.QHBoxLayout()
        self.filetable  = QtWidgets.QTableWidget()
        headers = ['Filename','Bytes written','Packets written','Date created','Date closed','Full filepath']
        self.filetable.setColumnCount(len(headers))
        self.filetable.setHorizontalHeaderLabels(headers)
        self.filetable.resizeColumnsToContents()
        self.filelab= QtWidgets.QLabel("File: ")
        self.byteslab   = QtWidgets.QLabel("Bytes written: ")
        self.packetslab = QtWidgets.QLabel("Packets written: ")
        # Table that displays all datastreams and the format as it is written to the file
        self.deviceinfoQtree = redvypr.gui.dictQTreeWidget(dataname='file status', show_datatype = False)
        # Update layout
        updatelayout = QtWidgets.QHBoxLayout()
        self.update_auto = QtWidgets.QCheckBox('Autoupdate file status')
        self.update_auto.setChecked(True)
        updatelayout.addWidget(self.update_auto)
        self.update_show_all_files = QtWidgets.QRadioButton('Show all files')
        #self.update_show_all_files.setChecked(True)
        self.update_show_cur_file = QtWidgets.QRadioButton('Show current file')
        self.update_show_cur_file.setChecked(True)
        show_group=QtWidgets.QButtonGroup()
        show_group.addButton(self.update_show_all_files)
        show_group.addButton(self.update_show_cur_file)
        self.show_group = show_group
        updatelayout.addWidget(self.update_show_all_files)
        updatelayout.addWidget(self.update_show_cur_file)

        self.update_show_full = QtWidgets.QRadioButton('Show all information')
        #self.update_show_full.setChecked(True)
        self.update_show_red = QtWidgets.QRadioButton('Show reduced information')
        self.update_show_red.setChecked(True)
        showred_group = QtWidgets.QButtonGroup()
        showred_group.addButton(self.update_show_full)
        showred_group.addButton(self.update_show_red)
        self.showred_group = showred_group
        updatelayout.addWidget(self.update_show_full)
        updatelayout.addWidget(self.update_show_red)



        hlayout.addWidget(self.byteslab)
        hlayout.addWidget(self.packetslab)
        layout.addWidget(self.filelab)        
        layout.addLayout(hlayout)
        layout.addWidget(self.filetable)
        layout.addWidget(self.deviceinfoQtree)
        # Add update options
        layout.addLayout(updatelayout)
        #self.text.insertPlainText("hallo!")

    def update_qtreewidget(self):
        funcname = __name__ + '.update_qtreewidget()'
        logger.debug(funcname)
        if self.update_auto.isChecked():
            devinfo = self.device.get_device_info(address=self.device.address_str)

            print('Deviceinfo!!!!', devinfo)
            print('__________')
            #devinfo = self.device.get_device_info()
            #print('Deviceinfo 2!!!!', devinfo)

            filename = devinfo['_deviceinfo']['filename']

            if self.update_show_red.isChecked():
                print('Hallo show reduced is checked')
                file_status = devinfo['_deviceinfo']['file_status_reduced']
                print('File status', file_status)
            else:
                print('Hallo show full is checked')
                file_status = devinfo['_deviceinfo']['file_status']
                print('File status', file_status)

            if self.update_show_cur_file.isChecked():
                print('Hallo current file is checked')
                #file_status = file_status[filename]
                print('File status', file_status)
            else:
                print('Hallo is not checked')
                print('File status', file_status)

            print('hallo hallo',file_status)
            # Update the qtree
            # https://stackoverflow.com/questions/9364754/remembering-scroll-value-of-a-qtreewidget-in-pyqt?rq=3
            bar = self.deviceinfoQtree.verticalScrollBar()
            yScroll = bar.value()
            print('File status',file_status)
            self.deviceinfoQtree.reload_data(file_status)
            self.deviceinfoQtree.verticalScrollBar().setSliderPosition(yScroll)

    def update_data(self,data):
        funcname = __name__ + '.update()'
        print(funcname,data)
        try:
            data['_deviceinfo']
        except:
            return

        # Update qtree
        try:
            # Test if the file status has changed, if yes make an update
            file_status_tmp = data['_deviceinfo']['file_status']
            try:
                self.update_qtreewidget()
            except:
                pass
                #logger.debug(funcname, exc_info=True)

        except:
            pass
            #logger.info(funcname, exc_info=True)


        try:
            #print('data',data)
            try:
                filename_table = self.filetable.item(0,0).text()
            except:
                filename_table = ''

            try:
                data['_deviceinfo']['filename']
                FLAG_FILEUPDATE = True
            except:
                FLAG_FILEUPDATE = False

            try:
                data['_deviceinfo']['csvcolumns']
                FLAG_CSVCOLUMNUPDATE = True
            except:
                FLAG_CSVCOLUMNUPDATE = False

            if FLAG_CSVCOLUMNUPDATE:
                #print('csv column update')
                csvcolumns = data['_deviceinfo']['csvcolumns']
                nrows = len(csvcolumns)
                self.datastreamtable.setRowCount(nrows)
                iaddr = self.datastreamtable_columns.index('Address')
                idatatype = self.datastreamtable_columns.index('Datatype')
                idatakey = self.datastreamtable_columns.index('Datakey')
                idevice = self.datastreamtable_columns.index('Device')
                iformat = self.datastreamtable_columns.index('Format')
                icolnr = self.datastreamtable_columns.index('Columnnr.')
                #self.datastreamtable_columns = ['Columnnr.', 'Datakey', 'Device', 'Datatype', 'Format', 'Address']

                for i,c in enumerate(csvcolumns):
                    item = QtWidgets.QTableWidgetItem(str(i))
                    self.datastreamtable.setItem(i, icolnr, item)

                    datatype = data['_deviceinfo']['header_datatype'][i]
                    item = QtWidgets.QTableWidgetItem(datatype)
                    self.datastreamtable.setItem(i, idatatype, item)

                    datakey = data['_deviceinfo']['header_datakeys'][i]
                    item = QtWidgets.QTableWidgetItem(datakey)
                    self.datastreamtable.setItem(i, idatakey, item)

                    dataformat = data['_deviceinfo']['csvformat'][i]
                    item = QtWidgets.QTableWidgetItem(dataformat)
                    self.datastreamtable.setItem(i, iformat, item)

                    datadevice = data['_deviceinfo']['header_device'][i]
                    item = QtWidgets.QTableWidgetItem(datadevice)
                    self.datastreamtable.setItem(i, idevice, item)

                    item = QtWidgets.QTableWidgetItem(str(c))
                    self.datastreamtable.setItem(i, iaddr, item)

                self.datastreamtable.resizeColumnsToContents()
            if FLAG_FILEUPDATE:
                #print('Filename table',filename_table,data['_deviceinfo']['filename'])
                try:
                    tclose = datetime.datetime.fromtimestamp(data['_deviceinfo']['closed']).strftime(
                        '%d-%m-%Y %H:%M:%S')
                    item = QtWidgets.QTableWidgetItem(tclose)
                    self.filetable.setItem(0, 4, item)
                except:
                    pass

                if filename_table != data['_deviceinfo']['filename']:
                    self.filetable.insertRow(0)
                    try:
                        #headers = ['Filename', 'Date created', 'Bytes written', 'Date closed']
                        item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename'])
                        self.filetable.setItem(0,0,item)
                        item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename_full'])
                        self.filetable.setItem(0, 5, item)
                        tcreate = datetime.datetime.fromtimestamp(data['_deviceinfo']['created']).strftime('%d-%m-%Y %H:%M:%S')
                        item = QtWidgets.QTableWidgetItem(tcreate)
                        self.filetable.setItem(0, 3, item)
                    except Exception as e:
                        logger.exception(e)

                item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['bytes_written']))
                self.filetable.setItem(0, 1, item)
                item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['packets_written']))
                self.filetable.setItem(0, 2, item)
                self.filetable.resizeColumnsToContents()
                self.filelab.setText("File: {:s}".format(data['_deviceinfo']['filename']))
                self.byteslab.setText("Bytes written: {:d}".format(data['_deviceinfo']['bytes_written']))
                self.packetslab.setText("Packets written: {:d}".format(data['_deviceinfo']['packets_written']))
        except:
            logger.info(funcname, exc_info=True)


        #self.text.insertPlainText(str(data['_deviceinfo']['data']))
        



import redvypr.data_packets

data_packet = {'data': 'Hello World!', 'device': 'test_device_0', 'host': {'hostname': 'redvypr', 'tstart': 1670422195.4220772, 'addr': '192.168.178.26', 'uuid': '20221207150955.421856-93328248922693-013', 'local': True}, 't': 1670422215.8423817, 'numpacket': 3}
print('Expanding address strings:')
print(redvypr.data_packets.expand_address_string('test'))
print(redvypr.data_packets.expand_address_string('data/randdata_1:redvypr@192.168.178.26::04283d40-ef3c-11ec-ab8f-21d63600f1d0'))
print(redvypr.data_packets.expand_address_string('data/randdata_1::04283d40-ef3c-11ec-ab8f-21d63600f1d0'))
print(redvypr.data_packets.expand_address_string('data:test/randata'))


d = []
d.append('data/rand1:randredvypr')
d.append('data/rand1')
d.append('t/randdata::5c9295ee-6f8e-11ec-9f01-f7ad581789cc')
d.append('?data/randdata_1:redvypr@192.168.178.26::04283d40-ef3c-11ec-ab8f-21d63600f1d0')
d.append('data/randdata_1:redvypr@192.168.178.26::04283d40-ef3c-11ec-ab8f-21d63600f1d0')
d.append('data/randdata_1:redvypr@192.168.178.26')
d.append('data/randdata_1::04283d40-ef3c-11ec-ab8f-21d63600f1d0')
d.append('data/randdata_1:redvypr')
d.append('data/randdata_1::99283d40-ef3c-11ec-ab8f-21d63600f1d0')
d.append('*/randdata_1::04283d40-ef3c-11ec-ab8f-21d63600f1d0')
d.append('data/randdata_1:*')
d.append('foo')
d.append('wrong number::of::@@::')




for i,datastream in enumerate(d):
    print('i',i)
    print('datastream orig:\t', datastream)
    d_expanded = redvypr.data_packets.expand_address_string(datastream)
    d_parsed = redvypr.data_packets.parse_addrstr(datastream)    
    print('datastream expanded:\t',d_expanded)
    print('datastream parsed:\t',d_parsed)    
    print('-----')
    print('-----')


# Test the address
addr = redvypr.data_packets.redvypraddress('data/randdata_1:redvypr@192.168.178.26::04283d40-ef3c-11ec-ab8f-21d63600f1d0')
for stype in addr.strtypes:
    print('Address:',addr.get_str(stype))


addr2 = redvypr.data_packets.redvypraddress('')
print('strtpyes',addr2.get_strtypes())

addr3 = redvypr.data_packets.redvypr_address('test')
print(addr3)

print('Test if address is in data_packet')
#addr4 = redvypr.data_packets.redvypr_address('data/*::20221207150955.421856-93328248922693-013peter')
addr4 = redvypr.data_packets.redvypraddress('data/*')
inpacket = data_packet in addr4
print('Addr4',addr4)
print('in packet:',addr4,inpacket)












from redvypr.devices.sensors.generic_sensor.generic_sensor import BinarySensor
from redvypr import RedvyprAddress
import re

# HF (Heatflow)
HF_test1 = b'$FC0FE7FFFE1567E3,HF,00000431.3125,108,-0.000072,2774.364,2782.398,2766.746\n'
HF_test2 = b'$FC0FE7FFFE1567E3,HF,00054411.3125,13603,0.000037,2780.217,2786.642,2774.316\n'
HF_split = b'\$(?P<MAC>.+),HF,(?P<counter>[0-9.]+),(?P<np>[0-9]+),(?P<HF_V>[-,\+]*[0-9.]+),(?P<NTC_R>.*)\n'
HF_str_format = {'MAC':'str','counter':'float','np':'int','HF_V':'float','NTC_R':'array'}
HF_datakey_metadata = {'MAC':{'unit':'MAC64','description':'MAC of the sensor'},'np':{'unit':'counter'},'HF_V':{'unit':'Volt'},'NTC_R':{'unit':'Ohm'}}
HF_packetid_format = 'HF_{MAC}'
HF = BinarySensor(name='HF', regex_split=HF_split,
                       str_format=HF_str_format,
                       datakey_metadata=HF_datakey_metadata,
                       packetid_format=HF_packetid_format,
                       datastream=str(RedvyprAddress('/k:data')))


print('Testing sensor definition')
rematch = re.finditer(HF_split, HF_test2)
print(rematch)
for r in rematch:
    print(r)
    redict = r.groupdict()
    print('redict', redict)

print('Done')





import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import pyqtgraph
import qtawesome as qta
import redvypr.data_packets
#from redvypr.gui import configWidget
from redvypr.devices.plot.plot_widgets import redvypr_numdisp_widget, redvypr_graph_widget, config_template_numdisp, config_template_graph
import redvypr.files as files
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
#from redvypr.redvypr_packet_statistic import do_data_statistics, create_data_statistic_dict
#from redvypr.configdata import configdata, getdata

_logo_file = files.logo_file
_icon_file = files.icon_file
pyqtgraph.setConfigOption('background', 'w')
pyqtgraph.setConfigOption('foreground', 'k')

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('plot')
logger.setLevel(logging.DEBUG)

description = 'Device that plots the received data'
config_template = {}
config_template['template_name']  = 'plot'
config_template['plots'] = {'type': 'list', 'modify': True, 'options': [config_template_numdisp, config_template_graph]}
config_template['dt_update'] = {'type':'float','default':0.25}
config_template['nx'] = {'type':'int','default':7}
config_template['ny'] = {'type':'int','default':6}
config_template['redvypr_device'] = {}
config_template['redvypr_device']['publish'] = False
config_template['redvypr_device']['subscribe'] = True
config_template['redvypr_device']['description'] = description
redvypr_devicemodule = True

def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    while True:
        time.sleep(0.05)
        while (datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
            except:
                data = None

            if (data is not None):
                command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
                # logger.debug('Got a command: {:s}'.format(str(data)))
                if (command is not None):
                    logger.debug('Command is for me: {:s}'.format(str(command)))
                    if(command == 'stop'):
                        logger.info(funcname + ': Stopped')
                        return

                #print('Got data')
                dataqueue.put(data)  # This has to be done, otherwise the gui does not get any data ...


class Device(RedvyprDevice):
    def __init__(self, **kwargs):
        """
        """
        super(Device, self).__init__(**kwargs)
        self.start = start

    def thread_start(self):
        """

        Returns:

        """
        self.connect_devices()
        super(Device,self).thread_start()

    def connect_devices(self):
        """ Connects devices, if they are not already connected
        """
        funcname = self.__class__.__name__ + '.connect_devices():'
        logger.debug(funcname)
        print('Connecting devices')
        # Check of devices have not been added
        plot_devices = []
        for plot in self.custom_config['plots']:  # Loop over all plots
            print('Config', self.custom_config)
            print('plot',plot)
            if (str(getdata(plot['type'])).lower() == 'numdisp'):
                datastream = plot['datastream']
                if (len(datastream) > 0):
                    address = redvypr.data_packets.redvypraddress(datastream)
                    plot_devices.append(address)

            elif (str(plot['type']).lower() == 'graph'):
                for l in plot['lines']:  # Loop over all lines in a plot
                    if (len(l['x']) > 0) and (len(l['y']) > 0):
                        xaddress = redvypr.data_packets.redvypr_address(str(l['x']))
                        yaddress = redvypr.data_packets.redvypr_address(str(l['y']))
                        xaddress = l['x'].xaddr
                        yaddress = l['y'].yaddr
                        plot_devices.append(xaddress)
                        plot_devices.append(yaddress)

        if True:
            print('Plot devices',plot_devices,self.name)
            for name in plot_devices:
                logger.info(funcname + 'Subscribing to device {:s}'.format(str(name)))
                ret = self.subscribe_address(name.get_str(),force=True)
                if (ret == False):
                    logger.info(funcname + 'Device is already subscribed')
                elif (ret == True):
                    logger.info(funcname + 'Device subscribed')


#
#
#
#
class displayDeviceWidget(QtWidgets.QWidget):
    """ Widget is a wrapper for several plotting widgets (numdisp, graph) 
    This widget can be configured with a configuration dictionary 
    """

    def __init__(self, device=None,deviceinitwidget=None):
        funcname = __name__ + '.init()'
        super(QtWidgets.QWidget, self).__init__()
        self.deviceinitwidget = deviceinitwidget
        # Let the configuration only be done by here, not in the initwidget
        #self.deviceinitwidget.config_widget.configtree.setEnabled(False)
        self.config = device.custom_config
        self.layout = QtWidgets.QVBoxLayout(self)
        self.device = device
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.configwidget = QtWidgets.QWidget()  # The configuration widget
        self.configplotwidget = QtWidgets.QWidget(
            parent=self.configwidget)  # The configuration widget of the individual plots
        self.configplotlayout = QtWidgets.QVBoxLayout(self.configplotwidget)
        self.displaywidget = PlotGridWidget(configplotwidget=self.configplotwidget,device=self.device,displaywidget = self)
        self.all_plots = self.displaywidget.all_plots
        self.layout.addWidget(self.splitter)
        self.splitter.addWidget(self.configwidget)
        self.splitter.addWidget(self.displaywidget)
        self.splitter.setStretchFactor(0, 1)
        self.splitter.setStretchFactor(1, 10)
        self.databuf = []
        self.configlayout = QtWidgets.QGridLayout(self.configwidget)
        self.status = {}
        self.status['last_update'] = time.time()
        self.init_configwidget()
        # Remove the plots from the list, this is necessary because addPlot will add the plot to the list again
        plots = []
        for p in reversed(self.config['plots']):
            print('p',p)
            p2 = copy.deepcopy(p)
            plots.append(p2)
            self.config['plots'].remove(p)

        if len(plots) == 0:
            graph = self.displaywidget.create_plot_widget('graph')
            self.displaywidget.addPlot(graph, 0, 0, 6, 6)

        for p in reversed(plots):
            print('p',p)
            x      = p['location']['x']
            y      = p['location']['y']
            width  = p['location']['width']
            height = p['location']['height']
            plotconfig = p
            w = self.displaywidget.create_plot_widget(p['type'],config = plotconfig)
            print('Adding plot',w)
            self.displaywidget.addPlot(w, y, x, height, width)



    def init_configwidget(self):
        self.add_button = QtWidgets.QPushButton('Add Plot')
        self.add_button.clicked.connect(self.add_plot_clicked)
        self.add_button.setCheckable(True)
        self.addplot_combo = QtWidgets.QComboBox()  # (buticon, 'Plot')
        self.addplot_combo.addItem('Graph')
        buticon = qta.icon('mdi6.chart-bell-curve-cumulative')  # Plot
        self.addplot_combo.setItemIcon(0, buticon)
        self.addplot_combo.addItem('Numeric Display')
        buticon = qta.icon('mdi6.order-numeric-ascending')  # Numeric display
        self.addplot_combo.setItemIcon(1, buticon)

        self.addplot_combo.addItem('Test')
        self.addplot_combo.setEnabled(False)
        self.addplot_combo.currentTextChanged.connect(self.add_plot_combo_changed)

        self.mod_button = QtWidgets.QPushButton('Modify')
        self.mod_button.clicked.connect(self.mod_plot_clicked)
        self.mod_button.setCheckable(True)

        self.rem_button = QtWidgets.QPushButton('Remove plot')
        self.rem_button.clicked.connect(self.rem_plot_clicked)
        self.rem_button.setCheckable(True)

        self.commit_button = QtWidgets.QPushButton('Commit')
        self.commit_button.clicked.connect(self.commit_plot_clicked)
        # self.commit_button.setCheckable(True)


        self.configlayout.addWidget(self.addplot_combo, 0, 0,1,2)
        self.configlayout.addWidget(self.add_button, 1, 0)
        self.configlayout.addWidget(self.rem_button, 1, 1)
        self.configlayout.addWidget(self.mod_button, 2, 0,1,2)
        self.configlayout.addWidget(self.commit_button, 3, 0,1,2)
        self.configlayout.setRowStretch(self.configlayout.rowCount(), 1)
        #self.configlayout.addWidget(self.configplotwidget, 3, 0, 3, 2)
        # self.configlayout.setRowStretch()

    def add_plot_combo_changed(self):
        plottype = self.addplot_combo.currentText()
        print('Plottype', plottype)


    def add_plot_clicked(self):
        self.add_plot_combo_changed()

        if (self.add_button.isChecked()):
            self.addplot_combo.setEnabled(True)
            self.displaywidget.rubber_enabled = True
            # Disable other buttons
            self.mod_button.setChecked(False)
            self.displaywidget.flag_add_plot = self.add_button.isChecked()
            self.displaywidget.flag_mod_plot = False
            self.displaywidget.flag_rem_plot = False
            self.rem_button.setChecked(False)
            self.displaywidget.modPlotclicked(False)
        else:
            self.addplot_combo.setEnabled(False)
            self.displaywidget.rubber_enabled = False

    def rem_plot_clicked(self):
        if (self.rem_button.isChecked()):
            self.displaywidget.flag_rem_plot = True
            self.displaywidget.flag_add_plot = False
            self.displaywidget.flag_mod_plot = False
            self.displaywidget.rubber_enabled = False
            self.addplot_combo.setEnabled(False)
            self.add_button.setChecked(False)
            self.mod_button.setChecked(False)
        else:
            self.displaywidget.flag_rem_plot = False

        self.displaywidget.remPlotclicked(self.rem_button.isChecked())

    def mod_plot_clicked(self):
        if (self.mod_button.isChecked()):
            self.displaywidget.rubber_enabled = False
            self.displaywidget.flag_add_plot = False
            self.displaywidget.flag_rem_plot = False
            self.addplot_combo.setEnabled(False)
            self.add_button.setChecked(False)
            self.rem_button.setChecked(False)

        self.displaywidget.flag_mod_plot = self.mod_button.isChecked()
        self.displaywidget.modPlotclicked(self.mod_button.isChecked())

    def commit_plot_clicked(self):
        self.displaywidget.commit_clicked()

    def update_data(self, data):
        funcname = __name__ + '.update_data():'
        logger.debug(funcname)
        tnow = time.time()
        self.databuf.append(data)

        #print('got data', data)
        #print('status', self.status)
        #print('config', self.config)
        #print(tnow,self.status['last_update'])
        #print((tnow - self.status['last_update']))
        # print('statistics',self.device.statistics)
        # Only plot the data in intervals of dt_update length, this prevents high CPU loads for fast devices
        if True:
            update = (tnow - self.status['last_update']) > self.config['dt_update']
            #print('update update', update)
            if (update):
                #self.status['last_update'] = tnow
                #print('updating', update)
                try:
                    for data in self.databuf:
                        #print('data',data)
                        for plotdict in self.all_plots:
                            plot = plotdict['plot']
                            #print('Plot ...',plot,plot.update_plot)
                            plot.update_plot(data)

                    self.databuf = []

                except Exception as e:
                    logger.exception(e)
                    logger.debug(funcname + 'Exception:' + str(e))

#
#
#
#
#
class PlotGridWidget(QtWidgets.QWidget):
    """
    The widget of the grid where the single plots can be added by the user
    """
    def __init__(self, configplotwidget,device=None,displaywidget = None):
        super(QtWidgets.QWidget, self).__init__()
        self.layout = QtWidgets.QGridLayout(self)
        self.configplotwidget = configplotwidget
        if(device is not None):
            initwidget = device.deviceinitwidget
            self.configwidget_global = initwidget.config_widget # The configuration widget that is shown on the init widget
            self.configwidget_global.config_changed_flag.connect(self.config_changed)
        self.displaywidget = displaywidget # The widget that is actually shown in the tabulator
        self.device = device
        self.redvypr = device.redvypr
        self.rubber_enabled = False
        self.flag_add_plot = False
        self.flag_mod_plot = False
        self.flag_rem_plot = False
        self.__add_location__ = None
        self.nx = self.device.config['nx'].data
        self.ny = self.device.config['ny'].data
        for i in range(self.ny):
            self.layout.setRowStretch(i, 1)

        for i in range(self.nx):
            self.layout.setColumnStretch(i, 1)

        self.gridcells = []
        self.all_plots = []  # A list of all plots added to the grid
        for i in range(self.nx):
            for j in range(self.ny):
                b = PlotGridWidgetButton()
                # b = QtWidgets.QPushButton()
                b.setCheckable(True)
                # Store the location indices as extra attributes
                b.__i__ = i
                b.__j__ = j
                self.gridcells.append(b)
                # b.setEnabled(False)
                b.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
                self.layout.addWidget(b, j, i)
                if(i==0) and (j == 0):
                    b.resize_signal.connect(self.resize_all_rubberbands)

        #testw = RandomDataWidget()
        #testw = redvypr_numdisp_widget()
        # self.addPlot(testw, 1, 1, 2, 2)
        #testg = redvypr_graph_widget()
        #self.addPlot(testg, 0, 3, 5, 3)


        self.rubberband = QtWidgets.QRubberBand(
            QtWidgets.QRubberBand.Rectangle, self)

        self.setMouseTracking(True)

    def mousePressEvent(self, event):
        self.origin = event.pos()
        if (self.rubber_enabled):
            self.rubberband.setGeometry(
                QtCore.QRect(self.origin, QtCore.QSize()))
            self.rubberband.show()
        QtWidgets.QWidget.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        if self.rubberband.isVisible():
            self.rubberband.setGeometry(
                QtCore.QRect(self.origin, event.pos()).normalized())

            print(self.origin)
            print(
                QtCore.QRect(self.origin, event.pos()).normalized())
        QtWidgets.QWidget.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):

        if self.rubberband.isVisible():  # New plot to be added
            self.rubberband.hide()
            selected = []

            rect = self.rubberband.geometry()
            FLAG_CHECKED = False
            FLAG_ALL_CHECKED = True

            for child in self.findChildren(QtWidgets.QPushButton):
                if rect.intersects(child.geometry()):
                    if (child.isChecked()):
                        FLAG_CHECKED = True
                    else:
                        FLAG_ALL_CHECKED = False

                    selected.append(child)
            for child in selected:
                if ((child.isChecked()) and (FLAG_CHECKED == False)) or FLAG_ALL_CHECKED:
                    child.setChecked(False)
                else:
                    child.setChecked(True)

        if (self.flag_rem_plot):  # If we wantto remove a widget
            print('Remove click')

        QtWidgets.QWidget.mouseReleaseEvent(self, event)

    def get_selected_index(self):
        """
        Gets the index of all selected gridbuttons (used for adding a device)
        Returns:

        """
        # Do we have an area larger than zero?
        # Adding the new widget
        iall = []
        jall = []
        for child in self.findChildren(QtWidgets.QPushButton):
            if child.isChecked():
                iall.append(child.__i__)
                jall.append(child.__j__)

        if (len(iall) > 0 and len(jall) > 0):
            inew = min(iall)
            di = max(iall) - min(iall) + 1
            jnew = min(jall)
            dj = max(jall) - min(jall) + 1
            self.__add_location__ = [jnew, inew, dj, di]
        else:
            self.__add_location__ = None

    def remPlotclicked(self, enabled):
        """
        Modify the existing plots

        Returns:

        """
        logger.debug('Removing')
        self.reset_all_rubberbands()
        if (enabled):  # Adding rubberbands to all plots
            self.show_all_rubberbands()
        else:  # Hiding all rubberbands
            self.hide_all_rubberbands()

    def modPlotclicked(self, enabled):
        """
        Modify the existing plots

        Returns:

        """
        logger.debug('Modifying')
        self.reset_all_rubberbands()
        if (enabled):  # Adding rubberbands to all plots
            self.show_all_rubberbands()
        else:  # Hiding all rubberbands
            self.hide_all_rubberbands()

    def show_all_rubberbands(self):
        for d in self.all_plots:
            w = d['plot']
            # rubberband = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)
            try:
                rubberband = d['rubber']
            except:
                rubberband = ResizableRubberBand(self)
                rubberband.mouse_pressed_right.connect(self.rubberband_clicked)
                rubberband.__config_widget__ = d['config']  # Add the configuration widget
                col = QtGui.QPalette()
                col.setBrush(QtGui.QPalette.Highlight, QtGui.QBrush(QtCore.Qt.red))
                rubberband.setPalette(col)
                rubberband.setWindowOpacity(.5)
                # print('Size', b, b.pos().x(), b.pos().y(), b.size())
                # print('Size', self, self.pos(), self.size())
                d['rubber'] = rubberband

            rubberband.setGeometry(QtCore.QRect(w.pos(), w.size()).normalized())
            rubberband.show()

    def hide_all_rubberbands(self):
        for d in self.all_plots:
            try:
                r = d['rubber']
                r.hide()
            except Exception as e:
                pass

    def reset_all_rubberbands(self):
        for d in self.all_plots:
            try:
                r = d['rubber']
                w = d['plot']
                col = QtGui.QPalette()
                col.setBrush(QtGui.QPalette.Highlight, QtGui.QBrush(QtCore.Qt.red))
                # r.setGeometry(QtCore.QRect(w.pos(), w.size()).normalized())
                r.setPalette(col)
                r.setWindowOpacity(.5)
                r.flag_rem_plot = False
                print('Reset done')
            except Exception as e:
                pass

    def resize_all_rubberbands(self):
        for d in self.all_plots:
            try:
                r = d['rubber']
                w = d['plot']
                r.setGeometry(QtCore.QRect(w.pos(), w.size()).normalized())
            except Exception as e:
                pass

    def rubberband_clicked(self):
        funcname = self.__class__.__name__ + '.rubberband_clicked'
        logger.debug(funcname)
        rubberband = self.sender()
        configplotwidget = rubberband.__config_widget__
        if True:
            configplotwidget.show()


    def create_plot_widget(self,plottype,config=None):
        """
        Creates a plot widget of type plottype and returns the widget
        Args:
            plottype:

        Returns:

        """
        funcname = __name__ + '.create_plot_widget():'

        self.displaywidget.__add_plottype__ = plottype
        if ('random' in plottype.lower()):
            plotwidget_ = RandomDataWidget
        elif ('num' in plottype.lower()):
            plotwidget_ = redvypr_numdisp_widget
        elif ('graph' in plottype.lower()):
            plotwidget_ = redvypr_graph_widget
        else:
            print('Not implemented yet: {:s}'.format(plottype))

        plotwidget = plotwidget_(config=config) # Call the plotwidget
        config_widget = configWidget(config=plotwidget.config, loadsavebutton=False, redvypr_instance=self.redvypr)
        config_widget.setWindowIcon(QtGui.QIcon(_icon_file))
        config_widget.config_changed_flag.connect(
            self.config_changed)  # whenever the configuration was changed, apply it for all plots with config_changed
        config_widget.plotwidget = plotwidget
        # Set the size
        config_widget.resize(1000, 800)  # TODO, calculate the size of the widget
        # plotwidget.config = config_widget.config
        plotwidget.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        # Add the configuration widget to the plotwidget
        plotwidget.config_widget = config_widget
        # Add a clear buffer button to the config widget if its a graph
        if ('graph' in plottype.lower()):
            config_widget.clearbtn = QtWidgets.QPushButton('Clear buffer')
            config_widget.layout.addWidget(config_widget.clearbtn, 1, 0,1,2)
            config_widget.clearbtn.clicked.connect(plotwidget.clear_buffer)

        return plotwidget



    def addPlot(self, plotwidget, j, i, height, width):
        """
        Adds a plot to the gridwidget.
        Args:
            plotwidget: The widget to be added
            j:
            i:
            height:
            width:

        Returns:

        """
        self.layout.addWidget(plotwidget, j, i, height, width)
        # Add the location information
        configuration = plotwidget.custom_config
        configuration['location']['x'].data = i
        configuration['location']['y'].data = j
        configuration['location']['height'].data = height
        configuration['location']['width'].data = width
        configuration.config_widget = plotwidget.config_widget
        configuration.plotwidget = plotwidget
        print('Added plot',type(configuration))
        self.device.config['plots'].append(configuration)
        d = {'plot': plotwidget, 'config': plotwidget.config_widget}
        self.all_plots.append(d)
        self.config_changed()
        plotwidget.show()
        
    def config_changed(self):
        """
        Function is called whenever the configuration has been changed. It updates the configuration of the plotwidgets
        and checkes the subscriptions


        Args:

        Returns:

        """
        funcname = self.__class__.__name__ + '.config_changed'
        configwidget = self.sender()

        if (configwidget is not self.configwidget_global):  # Reload the configwidget, if its not the sender
            self.configwidget_global.reload_config()

        if True:
            for p in self.device.config['plots']:
                try:
                    configwidget_tmp = p.config_widget
                except:
                    configwidget_tmp = None
                if(configwidget_tmp is not self.configwidget_global): # Reload the configwidget, if its not the sender
                    try:
                        p.config_widget.reload_config()
                    except:
                        pass

                try:
                    p.plotwidget.apply_config()
                except Exception as e:
                    logger.exception(e)

        #(re)connecting the devices
        self.device.connect_devices()

    def remPlot(self, plotwidget):
        """
        Removes a plotwidget from the grid

        Args:
            plotwidget:

        Returns:

        """
        self.layout.removeWidget(plotwidget)
        for d in self.all_plots:
            if (d['plot'] == plotwidget):
                print('removing from list')
                self.all_plots.remove(d)
                self.device.config['plots'].remove(plotwidget.custom_config)
                try:
                    r = d['rubber']
                    r.close()
                except:
                    pass
                break

        plotwidget.close()

    def remAllPlots(self):
        """
        Removes all plotwidget from the grid

        Args:

        Returns:

        """

        for d in reversed(self.all_plots):
            plotwidget = d['plot']
            print('removing from list')
            self.all_plots.remove(d)
            self.layout.removeWidget(plotwidget)
            self.device.config['plots'].remove(plotwidget.custom_config)
            try:
                r = d['rubber']
                r.close()
            except:
                pass

            plotwidget.close()


    def commit_clicked(self):
        """

        Returns:

        """
        logger.debug('Commit')
        # Plots have been moved
        if (self.flag_mod_plot):  # Flags are set by the gridwidget buttons
            for d in self.all_plots:
                try:
                    r = d['rubber']
                    plotwidget = d['plot']
                except Exception as e:
                    print('ohoh', e)
                    continue

                iall = []
                jall = []
                rect = r.geometry()
                for child in self.gridcells:
                    if rect.intersects(child.geometry()):
                        iall.append(child.__i__)
                        jall.append(child.__j__)

                if (len(iall) > 0 and len(jall) > 0):
                    inew = min(iall)
                    di = max(iall) - min(iall) + 1
                    jnew = min(jall)
                    dj = max(jall) - min(jall) + 1
                    print('d', d['plot'].custom_config)
                    # Update the location of the configuration
                    d['plot'].custom_config['location']['x'].data = inew
                    d['plot'].custom_config['location']['y'].data = jnew
                    d['plot'].custom_config['location']['width'].data = di
                    d['plot'].custom_config['location']['height'].data = dj
                    self.layout.removeWidget(plotwidget)
                    self.layout.addWidget(plotwidget, jnew, inew, dj, di)

        # Plots have been added
        if (self.flag_add_plot):  # Flags are set by the gridwidget buttons
            self.get_selected_index()  # update self.__add_location__
            if self.__add_location__ is not None:
                plottype = self.displaywidget.addplot_combo.currentText()
                addwidget_called = self.create_plot_widget(plottype)
                self.addPlot(addwidget_called, self.__add_location__[0], self.__add_location__[1],
                             self.__add_location__[2], self.__add_location__[3])

                # Remove all selected indices
                self.unselect_all()

            else:
                logger.debug('Not a valid location for adding plot')

        # Plots have been removed
        if (self.flag_rem_plot):  # Flags are set by the gridwidget buttons
            for d in reversed(self.all_plots):
                try:
                    r = d['rubber']
                    if r.flag_rem:  # If the remove flag is set
                        self.remPlot(d['plot'])
                except Exception as e:
                    logger.debug('Exception {:s}'.format(str(e)))

        print('Calling config changed')
        self.config_changed()


    def unselect_all(self):
        for child in self.findChildren(QtWidgets.QPushButton):
            child.setChecked(False)


class PlotGridWidgetButton(QtWidgets.QPushButton):
    resize_signal = QtCore.pyqtSignal()  # Signal notifying resize

    def __init__(self):
        super(QtWidgets.QWidget, self).__init__()

    def resizeEvent(self, event):
        self.resize_signal.emit()

    def mousePressEvent(self, event):
        QtWidgets.QWidget.mousePressEvent(self.parent(), event)


class RandomDataWidget(QtWidgets.QWidget):
    def __init__(self):
        super(QtWidgets.QWidget, self).__init__()
        self.i = 0
        self.texts = ['Hello', 'redvypr', 'data']
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_status)
        self.statustimer.start(2000)
        self.label = QtWidgets.QLabel(self.texts[0])
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(self.label)
        self.config = {}
        self.config['random'] = 10
        self.config_template = {}
        self.config_template['random'] = {'type': 'int', 'default': 11}
        self.setStyleSheet("background-color:green;")

    def update_status(self):
        self.i += 1
        self.i = self.i % len(self.texts)
        self.label.setText(self.texts[self.i])


class ResizableRubberBand(QtWidgets.QWidget):
    """Wrapper to make QRubberBand mouse-resizable using QSizeGrip

    Source: http://stackoverflow.com/a/19067132/435253
    """
    mouse_pressed_left = QtCore.pyqtSignal()  # Signal
    mouse_pressed_right = QtCore.pyqtSignal()  # Signal

    def __init__(self, parent):
        # super(Device, self).__init__(**kwargs)
        super(ResizableRubberBand, self).__init__(parent)

        self.setWindowFlags(QtCore.Qt.SubWindow)
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.grip1 = QtWidgets.QSizeGrip(self)
        self.grip2 = QtWidgets.QSizeGrip(self)
        self.layout.addWidget(self.grip1, 0, QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        self.layout.addWidget(self.grip2, 0, QtCore.Qt.AlignRight | QtCore.Qt.AlignBottom)

        self.rubberband = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)
        self.rubberband.move(0, 0)
        self.rubberband.show()

        self.flag_rem = False  # Flag for removal
        self.show()

    def resizeEvent(self, event):
        self.rubberband.resize(self.size())

    def mousePressEvent(self, event):
        print('Mouse press')
        if event.button() == QtCore.Qt.LeftButton:
            print("Left Button Clicked")
        elif event.button() == QtCore.Qt.RightButton:
            #do what you want here
            print("Right Button Clicked")
            self.mouse_pressed_right.emit()

        self.oldPos = event.globalPos()


    def mouseMoveEvent(self, event):
        print('Move')
        delta = QtCore.QPoint(event.globalPos() - self.oldPos)
        self.move(self.x() + delta.x(), self.y() + delta.y())
        self.oldPos = event.globalPos()

    def mouseReleaseEvent(self, event):
        print('Mouse release')
        try:
            flag_rem = self.parent().flag_rem_plot
            print('remove', self.parent().flag_rem_plot)
        except:
            flag_rem = False
            print('noremove')

        if (flag_rem and self.flag_rem == False):
            col = QtGui.QPalette()
            col.setBrush(QtGui.QPalette.Highlight, QtGui.QBrush(QtCore.Qt.black))
            self.setPalette(col)
            self.flag_rem = True
        else:
            self.flag_rem = False
            col = QtGui.QPalette()
            col.setBrush(QtGui.QPalette.Highlight, QtGui.QBrush(QtCore.Qt.red))
            self.setPalette(col)

        # self.__config_widget__
        # self.oldPos = event.globalPos()



import ast
import os
import re
import logging
import sys
from PyQt5 import QtWidgets, QtCore, QtGui
import multiprocessing
import argparse
import signal
import yaml
# Import redvypr specific stuff
import redvypr
import redvypr.files as files
from redvypr.redvypr_main_widget import redvyprMainWidget
import faulthandler
faulthandler.enable()

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('redvypr_main')
logger.setLevel(logging.INFO)


# Windows icon fix
# https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
import ctypes
myappid = u'redvypr.redvypr.version'  # arbitrary string
try:
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
except:
    pass

_logo_file = files.logo_file
_icon_file = files.icon_file


def merge_configuration(redvypr_config=None):
    """
    Merges a list of configurations
    :param redvypr_config:
    :return:
    """
    funcname = "merge_configuration():"
    parsed_devices = []
    logger.debug(funcname)

    if (redvypr_config is not None):
        logger.debug(funcname + 'Configuration: ' + str(redvypr_config))
        if (type(redvypr_config) == str):
            redvypr_config = [redvypr_config]
    else:
        return False

    config_tmp = redvypr.RedvyprConfig()
    devices_all = []
    devicepath_all = []
    for iconf, configraw in enumerate(redvypr_config):
        #print('Configraw',configraw)
        #print('iconf', iconf,type(configraw))
        if isinstance(configraw, redvypr.RedvyprConfig):
            logger.info(funcname + ' Found redvypr config')
            config_tmp = configraw
        elif (type(configraw) == str):
            logger.info(funcname + 'Opening yaml file: ' + str(configraw))
            if (os.path.exists(configraw)):
                fconfig = open(configraw)
                try:
                    config_tmp = yaml.load(fconfig, Loader=yaml.SafeLoader)
                except:
                    logger.warning(funcname + 'Could not load yaml file with safe loader')
                    fconfig.close()
                    fconfig = open(configraw)
                    try:
                        config_tmp = yaml.load(fconfig, Loader=yaml.CLoader)
                        logger.debug('Config tmp {}'.format(config_tmp))
                    except:
                        logger.warning(funcname + ' Could not load yaml file with x loader')
                        continue

                config_tmp = redvypr.RedvyprConfig(**config_tmp)
            else:
                logger.warning(funcname + 'Yaml file: ' + str(configraw) + ' does not exist!')
                continue
        elif (type(configraw) == dict):
            logger.debug(funcname + 'Opening dictionary')
            config_tmp = redvypr.RedvyprConfig(**configraw)
        else:
            logger.warning(funcname + 'Unknown type of configuration {:s}'.format(type(configraw)))
            continue

        # Merge the configuration into one big dictionary
        devices_all.extend(config_tmp.devices)
        devicepath_all.extend(config_tmp.devicepaths)
        #print('Config tmp', config_tmp)
        # config = config.model_copy(update=config_tmp)
    config_tmp2 = redvypr.RedvyprConfig(devices=devices_all, devicepaths=devicepath_all)
    config = config_tmp2.model_copy(update=config_tmp.model_dump(exclude=['devices', 'devicepaths']))
    #print('Config', config)
    return config

def split_quotedstring(qstr, separator=','):
    """ Splits a string
    """
    r = re.compile("'.+?'")  # Single quoted string

    d = qstr[:]
    quoted_list = r.findall(d)
    quoted_dict = {}
    for fstr in quoted_list:
        u1 = uuid.uuid4()
        d = d.replace(fstr, u1.hex, 1)
        quoted_dict[u1.hex] = fstr

    ds = d.split(separator)
    for i, dpart in enumerate(ds):
        for k in quoted_dict.keys():
            dpart = dpart.replace(k, quoted_dict[k])
            ds[i] = dpart

    return ds

#
#
# Main function called from os
#
#
#
def redvypr_main():
    redvypr_help = 'redvypr'
    config_help_verbose = 'verbosity, if argument is called at least once loglevel=DEBUG, otherwise loglevel=INFO'
    config_help = 'Using a yaml config file'
    config_help_nogui = 'start redvypr without a gui'
    config_help_path = 'add path to search for redvypr modules'
    config_help_hostname = 'hostname of redvypr, overwrites the hostname in a possible configuration '
    add_device_example = '\t-a test_device, s, [mp / th], loglevel: [DEBUG / INFO / WARNING], name: test_1, subscribe: "*"'
    add_device_example_2 = ', also device specific configuration can be set similarly: -a test_device,delay_s: 0.4, '
    config_help_add = 'add device, can be called multiple times, optional options/configuration can be added by comma separated input:' + add_device_example + add_device_example_2
    config_help_list = 'lists all known devices'
    config_optional = 'optional information about the redvypr instance, multiple calls possible or separated by ",". Given as a key:data pair: --hostinfo location:lab --hostinfo lat:10.2,lon:30.4. The data is tried to be converted to an int, if that is not working as a float, if that is neither working at is passed as string'
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', '-v', action='count', help=config_help_verbose)
    parser.add_argument('--config', '-c', help=config_help)
    parser.add_argument('--nogui', '-ng', help=config_help_nogui, action='store_true')
    parser.add_argument('--add_path', '-p', help=config_help_path)
    parser.add_argument('--hostname', '-hn', help=config_help_hostname)
    parser.add_argument('--metadata', '-m', help=config_optional, action='append')
    parser.add_argument('--add_device', '-a', help=config_help_add, action='append')
    parser.add_argument('--list_devices', '-l', help=config_help_list, action='store_true')
    parser.set_defaults(nogui=False)
    args = parser.parse_args()

    # Check if list devices only
    if (args.list_devices):
        # Set the nogui flag
        args.nogui = True

    logging_level = logging.INFO
    if (args.verbose == None):
        logging_level = logging.INFO
        loglevel_redvypr=None
    elif (args.verbose >= 1):
        print('Debug logging level')
        logging_level = logging.DEBUG
        loglevel_redvypr = 'DEBUG'

    logger.setLevel(logging_level)

    # Check if we have a redvypr.yaml, TODO, add also default path
    config_all = [] # Make a config all, the list can have several dictionaries that will be all processed by the redvypr initialization
    if (os.path.exists('redvypr.yaml')):
        config_all.append('redvypr.yaml')

    # Adding device module pathes
    if (args.add_path is not None):
        # print('devicepath',args.add_path)
        modpath = os.path.abspath(args.add_path)
        # print('devicepath',args.add_path,modpath)
        # print('Modpath',modpath)
        config_add = redvypr.RedvyprConfig(devicepaths=[modpath])
        config_all.append(config_add)

    # Add the configuration
    config = args.config
    if (config is not None):
        config_all.append(config)

    # Add device
    if (args.add_device is not None):
        devices_add = []
        #print('devices!', args.add_device)
        for d in args.add_device:
            deviceconfig = redvypr.RedvyprDeviceConfig().model_dump()
            deviceconfig['custom_config'] = {} # Change the None manually to a dictionary
            #deviceconfig = {'base_config':{},'config':{},'subscriptions':[]}
            #deviceconfig['base_config'] = {'autostart':False,'loglevel':logging_level,'multiprocess':'qthread'}
            if(',' in d):
                logger.debug('Found options')
                #devicemodulename = d.split(',')[0]
                #options = d.split(',')[1:]
                # Split the string, using csv reader to have quoted strings conserved
                options_all = split_quotedstring(d)
                #print('options all',options_all)
                devicemodulename = options_all[0]
                options = options_all[1:]
                #print('options', options,type(options))
                for indo,option in enumerate(options):
                    #print('Option',option,len(option),indo)
                    if(option == 's'):
                        print('Autostart')
                        deviceconfig['base_config']['autostart'] = True
                    elif (option == 'mp' or option == 'multiprocess') and (':' not in option):
                        deviceconfig['base_config']['multiprocess'] = 'multiprocess'
                    elif (option == 'th' or option == 'thread') and (':' not in option):
                        deviceconfig['base_config']['multiprocess'] = 'qthread'
                    elif (':' in option):
                        key = option.split(':')[0]
                        data = option.split(':')[1]
                        #print('data before',data,key)
                        if (data[0] == "'") and (data[-1] == "'"):
                            try:
                                data = ast.literal_eval(data[1:-1])
                            except Exception as e:
                                logger.info('Error parsing options:', exc_info=True)
                        else:
                            try:
                                data = int(data)
                            except:
                                try:
                                    data = float(data)
                                except:
                                    pass

                        #print('Data', data)
                        #print('type Data', type(data))
                        print('Data', data)
                        print('key', key)
                        if(key == 'name'):
                            deviceconfig['base_config'][key] = data
                        elif (key == 'loglevel') or (key == 'll'):
                            try:
                                loglevel_tmp = data
                                loglevel_device = getattr(logging, loglevel_tmp.upper())
                            except Exception as e:
                                print(e)
                                loglevel_tmp = 'INFO'
                                loglevel_device = getattr(logging, loglevel_tmp.upper())

                            print('Setting device {:s} to loglevel {:s}'.format(devicemodulename,loglevel_tmp))
                            deviceconfig['base_config']['loglevel'] = loglevel_device
                        elif (key.lower() == 'subscribe'):
                            print('Add subscription {}'.format(str(data)))
                            deviceconfig['subscriptions'].append(data)
                        else:
                            print('Adding key',key,data)
                            deviceconfig['custom_config'][key] = data
            else:
                devicemodulename = d

            deviceconfig['devicemodulename'] = devicemodulename
            devconfig = redvypr.RedvyprDeviceConfig(**deviceconfig)
            devices_add.append(devconfig)
            logger.info('Adding device {}'.format(d))

        config_devices = redvypr.RedvyprConfig(devices=devices_add)
        print('Devices to add')
        print('D',config_devices)
        print('Done')
        config_all.append(config_devices)

    # Add hostname
    if (args.hostname is not None):
        hostname = args.hostname
    else:
        hostname = 'redvypr'

    # Add metadata
    if (args.metadata is not None):
        metadata = args.metadata
        # Add optional metadata
        metadata_tmp = {}
        for i in metadata:
            for info in i.split(','):
                #print('Info',info)
                if(':' in info):
                    key = info.split(':')[0]
                    data = info.split(':')[1]
                    try:
                        data = int(data)
                    except:
                        try:
                            data = float(data)
                        except:
                            pass

                    metadata_tmp[key] = data
                else:
                    logger.warning('Not a key:data pair in metadata, skipping {:sf}'.format(info))

        metadata_obj = redvypr.RedvyprMetadata(**metadata_tmp)
        config_metadata = redvypr.RedvyprConfig(metadata=metadata_obj)
        config_all.append(config_metadata)

    print('Config all',config_all)
    config = merge_configuration(config_all)
    print('Config',config)

    #config_all.append({'hostinfo_opt':hostinfo_opt})
    #print('Hostinfo', hostinfo)
    #print('Hostinfo opt', hostinfo_opt)

    logger.debug('Configuration:\n {:s}\n'.format(str(config_all)))
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
    # GUI oder command line?
    if (args.nogui):
        def handleIntSignal(signum, frame):
            '''Ask app to close if Ctrl+C is pressed.'''
            print('Received CTRL-C: Closing now')
            sys.exit()

        signal.signal(signal.SIGINT, handleIntSignal)
        app = QtCore.QCoreApplication(sys.argv)
        redvypr_obj = redvypr.Redvypr(config=config, hostname=hostname, nogui=True, loglevel=loglevel_redvypr)
        # Check if the devices shall be listed only
        if (args.list_devices):
            devices = redvypr_obj.get_known_devices()
            print('Known devices')
            for d in devices:
                print(d)

            sys.exit()
        sys.exit(app.exec_())
    else:
        app = QtWidgets.QApplication(sys.argv)
        app.setWindowIcon(QtGui.QIcon(_icon_file))
        screen = app.primaryScreen()
        # print('Screen: %s' % screen.name())
        size = screen.size()
        # print('Size: %d x %d' % (size.width(), size.height()))
        rect = screen.availableGeometry()
        width = int(rect.width() * 4 / 5)
        height = int(rect.height() * 2 / 3)

        logger.debug(
            'Available screen size: {:d} x {:d} using {:d} x {:d}'.format(rect.width(), rect.height(), width, height))
        ex = redvyprMainWidget(width=width, height=height, config=config, hostname=hostname, loglevel=loglevel_redvypr)

        sys.exit(app.exec_())


if __name__ == '__main__':
    #https://stackoverflow.com/questions/46335842/python-multiprocessing-throws-error-with-argparse-and-pyinstaller
    multiprocessing.freeze_support()
    redvypr_main()


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import serial
import serial.tools.list_ports
import logging
import sys
import pydantic
from redvypr.data_packets import check_for_command
#from redvypr.redvypr_packet_statistic import do_data_statistics, create_data_statistic_dict


description = 'Reading data from a serial device'


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('serial_single')
logger.setLevel(logging.DEBUG)

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Reads to and write from a serial device'
    gui_tablabel_display: str = 'Serial data'


class DeviceCustomConfig(pydantic.BaseModel):
    baud: int = 4800
    parity: int = serial.PARITY_NONE
    stopbits: int = serial.STOPBITS_ONE
    bytesize: int = serial.EIGHTBITS
    dt_poll: float = 0.05
    chunksize: int = pydantic.Field(default=1000, description='The maximum amount of bytes read with one chunk')
    packetdelimiter: str = pydantic.Field(default='\n', description='The delimiter to distinuish packets')
    comport: str = ''


config_template              = {}
config_template['comport']   = {'type':'str'}
config_template['baud']      = {'type':'int','default':4800}
config_template['parity']    = {'type':'int','default':serial.PARITY_NONE}
config_template['stopbits']  = {'type':'int','default':serial.STOPBITS_ONE}
config_template['bytesize']  = {'type':'int','default':serial.EIGHTBITS}
config_template['dt_poll']   = {'type':'float','default':0.05}
config_template['chunksize'] = {'type':'int','default':1000} # The maximum amount of bytes read with one chunk
config_template['packetdelimiter'] = {'type':'str','default':'\n'} # The maximum amount of bytes read with one chunk
config_template['redvypr_device'] = {}
config_template['redvypr_device']['publishes']   = True
config_template['redvypr_device']['subscribes']  = True
config_template['redvypr_device']['description'] = description
redvypr_devicemodule = True

def start(device_info, config={}, dataqueue=None, datainqueue=None, statusqueue=None):
    """

    """
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Starting reading serial data')
    chunksize   = config['chunksize'] #The maximum amount of bytes read with one chunk
    serial_name = config['comport']
    baud        = config['baud']
    parity      = config['parity']
    stopbits    = config['stopbits']
    bytesize    = config['bytesize']
    dt_poll     = config['dt_poll']

    print('Starting',config)
    
    newpacket   = config['packetdelimiter']
    # Check if a delimiter shall be used (\n, \r\n, etc ...)
    if(len(newpacket)>0):
        FLAG_DELIMITER = True
    else:
        FLAG_DELIMITER = False
    if(type(newpacket) is not bytes):
        newpacket = newpacket.encode('utf-8')
        
    rawdata_all    = b''
    dt_update      = 1 # Update interval in seconds
    bytes_read     = 0
    sentences_read = 0
    bytes_read_old = 0 # To calculate the amount of bytes read per second
    t_update       = time.time()
    serial_device = False
    if True:
        try:
            serial_device = serial.Serial(serial_name,baud,parity=parity,stopbits=stopbits,bytesize=bytesize,timeout=0)
            #print('Serial device 0',serial_device)
            #serial_device.timeout(0.05)
            #print('Serial device 1',serial_device)                        
        except Exception as e:
            #print('Serial device 2',serial_device)
            logger.debug(funcname + ': Exception open_serial_device {:s} {:d}: '.format(serial_name,baud) + str(e))
            return False

    got_dollar = False    
    while True:
        # TODO, here commands could be send as well
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if (data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            # logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                if command == 'stop':
                    serial_device.close()
                    sstr = funcname + ': Command is for me: {:s}'.format(str(command))
                    logger.debug(sstr)
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    return


        time.sleep(dt_poll)
        ndata = serial_device.inWaiting()
        try:
            rawdata_tmp = serial_device.read(ndata)
        except Exception as e:
            print(e)
            #print('rawdata_tmp', rawdata_tmp)

        nread = len(rawdata_tmp)
        if True:
            if nread > 0:
                bytes_read  += nread
                rawdata_all += rawdata_tmp
                #print('rawdata_all',rawdata_all)
                FLAG_CHUNK = len(rawdata_all) > chunksize
                if(FLAG_CHUNK):
                    data               = {'t':time.time()}
                    data['data']       = rawdata_all
                    data['comport']    = serial_device.name
                    data['bytes_read'] = bytes_read
                    dataqueue.put(data)
                    rawdata_all = b''

                # Check if the newpacket character in the data
                if(FLAG_DELIMITER):
                    FLAG_CHAR = newpacket in rawdata_all
                    if(FLAG_CHAR):
                        rawdata_split = rawdata_all.split(newpacket)
                        #print('rawdata_all', rawdata_all)
                        if(len(rawdata_split)>1): # If len==0 then character was not found
                            for ind in range(len(rawdata_split)-1): # The last packet does not have the split character
                                sentences_read += 1
                                raw = rawdata_split[ind] + newpacket # reconstruct the data
                                #print('raw', raw)
                                data               = {'t':time.time()}
                                data['data']       = raw
                                data['comport']    = serial_device.name
                                data['bytes_read'] = bytes_read
                                data['sentences_read'] = sentences_read
                                dataqueue.put(data)

                            rawdata_all = rawdata_split[-1]
        
            
            
        if((time.time() - t_update) > dt_update):
            dbytes = bytes_read - bytes_read_old
            bytes_read_old = bytes_read
            bps = dbytes/dt_update# bytes per second
            #print('ndata',len(rawdata_all),'rawdata',rawdata_all,type(rawdata_all))
            #print('bps',bps)
            t_update = time.time()
            
                





class initDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QVBoxLayout(self)
        self.device   = device
        self.serialwidget = QtWidgets.QWidget()
        self.init_serialwidget()
        self.label    = QtWidgets.QLabel("Serial device")
        #self.startbtn = QtWidgets.QPushButton("Open device")
        #self.startbtn.clicked.connect(self.start_clicked)
        #self.stopbtn = QtWidgets.QPushButton("Close device")
        #self.stopbtn.clicked.connect(self.stop_clicked)
        layout.addWidget(self.label)        
        layout.addWidget(self.serialwidget)
        #layout.addWidget(self.startbtn)
        #layout.addWidget(self.stopbtn)
        
    def init_serialwidget(self):
        """Fills the serial widget with content
        """
        layout = QtWidgets.QGridLayout(self.serialwidget)
        # Serial baud rates
        baud = [300,600,1200,2400,4800,9600,19200,38400,57600,115200,576000,921600]
        self._combo_serial_devices = QtWidgets.QComboBox()
        #self._combo_serial_devices.currentIndexChanged.connect(self._serial_device_changed)
        self._combo_serial_baud = QtWidgets.QComboBox()
        for b in baud:
            self._combo_serial_baud.addItem(str(b))

        self._combo_serial_baud.setCurrentIndex(4)
        # creating a line edit
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
  
        # setting line edit
        self._combo_serial_baud.setLineEdit(edit)
        
        self._combo_parity = QtWidgets.QComboBox()
        self._combo_parity.addItem('None')
        self._combo_parity.addItem('Odd')
        self._combo_parity.addItem('Even')
        self._combo_parity.addItem('Mark')
        self._combo_parity.addItem('Space')
        
        self._combo_stopbits = QtWidgets.QComboBox()
        self._combo_stopbits.addItem('1')
        self._combo_stopbits.addItem('1.5')
        self._combo_stopbits.addItem('2')
        
        self._combo_databits = QtWidgets.QComboBox()
        self._combo_databits.addItem('8')
        self._combo_databits.addItem('7')
        self._combo_databits.addItem('6')
        self._combo_databits.addItem('5')
        
        self._button_serial_openclose = QtWidgets.QPushButton('Open')
        self._button_serial_openclose.clicked.connect(self.start_clicked)


        # Check for serial devices and list them
        for comport in serial.tools.list_ports.comports():
            self._combo_serial_devices.addItem(str(comport.device))

        #How to differentiate packets
        self._packet_ident_lab = QtWidgets.QLabel('Packet identification')
        self._packet_ident     = QtWidgets.QComboBox()
        self._packet_ident.addItem('newline \\n')
        self._packet_ident.addItem('newline \\r\\n')
        self._packet_ident.addItem('None')
        # Max packetsize
        self._packet_size_lab   = QtWidgets.QLabel("Maximum packet size")
        self._packet_size_lab.setToolTip('The number of received bytes after which a packet is sent.\n Add 0 for no size check')
        onlyInt = QtGui.QIntValidator()
        self._packet_size       = QtWidgets.QLineEdit()
        self._packet_size.setValidator(onlyInt)
        self._packet_size.setText('0')
        #self.packet_ident

        layout.addWidget(self._packet_ident,0,1)
        layout.addWidget(self._packet_ident_lab,0,0)
        layout.addWidget(self._packet_size_lab,0,2)
        layout.addWidget(self._packet_size,0,3)
        layout.addWidget(QtWidgets.QLabel('Serial device'),1,0)
        layout.addWidget(self._combo_serial_devices,2,0)
        layout.addWidget(QtWidgets.QLabel('Baud'),1,1)
        layout.addWidget(self._combo_serial_baud,2,1)
        layout.addWidget(QtWidgets.QLabel('Parity'),1,2)  
        layout.addWidget(self._combo_parity,2,2) 
        layout.addWidget(QtWidgets.QLabel('Databits'),1,3)  
        layout.addWidget(self._combo_databits,2,3) 
        layout.addWidget(QtWidgets.QLabel('Stopbits'),1,4)  
        layout.addWidget(self._combo_stopbits,2,4) 
        layout.addWidget(self._button_serial_openclose,2,5)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)
        
    
    def update_buttons(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """

        status = self.device.get_thread_status()
        thread_status = status['thread_running']

        if(thread_status):
            self._button_serial_openclose.setText('Close')
            self._combo_serial_baud.setEnabled(False)
            self._combo_serial_devices.setEnabled(False)
        else:
            self._button_serial_openclose.setText('Open')
            self._combo_serial_baud.setEnabled(True)
            self._combo_serial_devices.setEnabled(True)
        
            
    def start_clicked(self):
        #print('Start clicked')
        button = self._button_serial_openclose
        #print('Start clicked:' + button.text())
        #config_template['comport'] = {'type': 'str'}
        #config_template['baud'] = {'type': 'int', 'default': 4800}
        #config_template['parity'] = {'type': 'int', 'default': serial.PARITY_NONE}
        #config_template['stopbits'] = {'type': 'int', 'default': serial.STOPBITS_ONE}
        #config_template['bytesize'] = {'type': 'int', 'default': serial.EIGHTBITS}
        #config_template['dt_poll'] = {'type': 'float', 'default': 0.05}
        #config_template['chunksize'] = {'type': 'int',
        #                                'default': 1000}  # The maximum amount of bytes read with one chunk
        #config_template['packetdelimiter'] = {'type': 'str',
        #                                      'default': '\n'}  # The maximum amount of bytes read with one chunk
        if('Open' in button.text()):
            button.setText('Close')
            serial_name = str(self._combo_serial_devices.currentText())
            serial_baud = int(self._combo_serial_baud.currentText())
            self.device.custom_config.comport = serial_name
            self.device.custom_config.baud = serial_baud
            stopbits = self._combo_stopbits.currentText()
            if(stopbits=='1'):
                self.device.custom_config.stopbits =  serial.STOPBITS_ONE
            elif(stopbits=='1.5'):
                self.device.custom_config.stopbits =  serial.STOPBITS_ONE_POINT_FIVE
            elif(stopbits=='2'):
                self.device.custom_config.stopbits =  serial.STOPBITS_TWO
                
            databits = int(self._combo_databits.currentText())
            self.device.custom_config.bytesize = databits

            parity = self._combo_parity.currentText()
            if(parity=='None'):
                self.device.custom_config.parity = serial.PARITY_NONE
            elif(parity=='Even'):                
                self.device.custom_config.parity = serial.PARITY_EVEN
            elif(parity=='Odd'):                
                self.device.custom_config.parity = serial.PARITY_ODD
            elif(parity=='Mark'):                
                self.device.custom_config.parity = serial.PARITY_MARK
            elif(parity=='Space'):                
                self.device.custom_config.parity = serial.PARITY_SPACE
                

            self.device.thread_start()
        else:
            self.stop_clicked()

    def stop_clicked(self):
        button = self._button_serial_openclose
        self.device.thread_stop()
        button.setText('Closing') 
        #self._combo_serial_baud.setEnabled(True)
        #self._combo_serial_devices.setEnabled(True)      



class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QVBoxLayout(self)
        hlayout        = QtWidgets.QHBoxLayout()
        self.device = device
        self.bytes_read = QtWidgets.QLabel('Bytes read: ')
        self.lines_read = QtWidgets.QLabel('Lines read: ')
        self.text     = QtWidgets.QPlainTextEdit(self)
        self.text.setReadOnly(True)
        self.text.setMaximumBlockCount(10000)
        hlayout.addWidget(self.bytes_read)
        hlayout.addWidget(self.lines_read)
        layout.addLayout(hlayout)
        layout.addWidget(self.text)

    def update_data(self,data):
        funcname = __name__ + '.update():'
        #print('data',data)
        try:
            bstr = "Bytes read: {:d}".format(data['bytes_read'])
            lstr = "Sentences read: {:d}".format(data['sentences_read'])
            self.bytes_read.setText(bstr)
            self.lines_read.setText(lstr)
        except Exception as e:
            logger.exception(e)
        try:
            self.text.insertPlainText(str(data['data']))
            self.text.insertPlainText('\n')
        except Exception as e:
            logger.debug(funcname,exc_info=True)
        



import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import logging
import sys
import yaml
import copy
import gzip
import os
import queue
import pydantic
import typing
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
from redvypr.packet_statistic import do_data_statistics, create_data_statistic_dict

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('rawdatawriter')
logger.setLevel(logging.DEBUG)
redvypr_devicemodule = True


class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = "Saves the raw redvypr packets into a file"
    gui_tablabel_display: str = 'Logging status'
    gui_icon: str = 'mdi.code-json'

class DeviceCustomConfig(pydantic.BaseModel):
    dt_sync: int = pydantic.Field(default=5, description='Time after which an open file is synced on disk')
    dt_newfile: int = pydantic.Field(default=60, description='Time after which a new file is created')
    dt_newfile_unit: typing.Literal['seconds', 'hours', 'days'] = pydantic.Field(default='seconds')
    dt_update: int = pydantic.Field(default= 5, description='Time after which an upate is sent to the gui')
    clearqueue: bool = pydantic.Field(default=True,
                                      description='Flag if the buffer of the subscribed queue should be emptied before start')
    size_newfile: int = pydantic.Field(default= 0, description='Size after which a new file is created')
    size_newfile_unit: typing.Literal['bytes', 'kB', 'MB'] = pydantic.Field(default='bytes')
    datafolder: str = pydantic.Field(default='./', description='Folder the data is saved to')
    fileextension: str = pydantic.Field(default='redvypr_yaml', description='File extension, if empty not used')
    fileprefix: str = pydantic.Field(default='redvypr', description='If empty not used')
    filepostfix: str = pydantic.Field(default='raw', description='If empty not used')
    filedateformat: str = pydantic.Field(default='%Y-%m-%d_%H%M%S', description='Dateformat used in the filename, must be understood by datetime.strftime')
    filecountformat: str = pydantic.Field(default='04', description='Format of the counter. Add zero if trailing zeros are wished, followed by number of digits. 04 becomes {:04d}')
    filegzipformat: str = pydantic.Field(default='gz', description='If empty, no compression done')


def create_logfile(config,count=0):
    funcname = __name__ + '.create_logfile():'
    logger.debug(funcname)

    filename = ''


    if(len(config['fileprefix'])>0):
        filename += config['fileprefix']

    if (len(config['filedateformat']) > 0):
        tstr = datetime.datetime.now().strftime(config['filedateformat'])
        filename += '_' + tstr

    if (len(config['filecountformat']) > 0):
        cstr = "{:" + config['filecountformat'] +"d}"
        filename += '_' + cstr.format(count)

    if (len(config['filepostfix']) > 0):
        filename += '_' + config['filepostfix']

    if (len(config['fileextension']) > 0):
        filename += '.' + config['fileextension']

    if (len(config['filegzipformat']) > 0):
        filename += '.' + config['filegzipformat']
        FLAG_GZIP = True
    else:
        FLAG_GZIP = False

    if len(config['datafolder']) > 0:
        if os.path.isdir(config['datafolder']):
            #filename += config['datafolder']
            filename = os.path.join(config['datafolder'],filename)
        else:
            logger.warning(funcname + ' Data folder {:s} does not exist.'.format(filename))
            return None

    logger.info(funcname + ' Will create a new file: {:s}'.format(filename))
    if FLAG_GZIP:
        try:
            f = gzip.open(filename,'w')
            logger.debug(funcname + ' Opened file: {:s}'.format(filename))
        except Exception as e:
            logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
            return None
    else:
        try:
            f = open(filename,'wb+')
            logger.debug(funcname + ' Opened file: {:s}'.format(filename))
        except Exception as e:
            logger.warning(funcname + ' Error opening file:' + filename + ':' + str(e))
            return None
       
    return [f,filename]

def start(device_info, config, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Opening writing:')
    #print('Config',config)
    if config['clearqueue']:
        while (datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
            except:
                break
    count = 0
    if True:
        try:
            dtneworig  = config['dt_newfile']
            dtunit     = config['dt_newfile_unit']
            if(dtunit.lower() == 'seconds'):
                dtfac = 1.0
            elif(dtunit.lower() == 'hours'):
                dtfac = 3600.0
            elif(dtunit.lower() == 'days'):
                dtfac = 86400.0
            else:
                dtfac = 0
                
            dtnews     = dtneworig * dtfac
            logger.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['dt_newfile'],config['dt_newfile_unit']))
        except Exception as e:
            logger.exception(e)
            dtnews = 0
            
        try:
            sizeneworig  = config['size_newfile']
            sizeunit     = config['size_newfile_unit']
            if(sizeunit.lower() == 'kb'):
                sizefac = 1000.0
            elif(sizeunit.lower() == 'mb'):            
                sizefac = 1e6
            elif(sizeunit.lower() == 'bytes'):            
                sizefac = 1 
            else:
                sizefac = 0
                
            sizenewb     = sizeneworig * sizefac # Size in bytes
            logger.info(funcname + ' Will create new file every {:d} {:s}.'.format(config['size_newfile'],config['size_newfile_unit']))
        except Exception as e:
            logger.exception(e)
            sizenewb = 0  # Size in bytes
            
            
    try:
        config['dt_sync']
    except:
        config['dt_sync'] = 5

    bytes_written = 0
    packets_written = 0
    bytes_written_total = 0
    packets_written_total = 0

    [f,filename] = create_logfile(config,count)
    data_stat = {'_deviceinfo': {}}
    data_stat['_deviceinfo']['filename'] = filename
    data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
    data_stat['_deviceinfo']['created'] = time.time()
    data_stat['_deviceinfo']['bytes_written'] = bytes_written
    data_stat['_deviceinfo']['packets_written'] = packets_written
    dataqueue.put(data_stat)
    count += 1
    statistics = create_data_statistic_dict()
    if(f == None):
       return None
    

    
    tfile           = time.time() # Save the time the file was created
    tflush          = time.time() # Save the time the file was created
    tupdate         = time.time() # Save the time for the update timing
    FLAG_RUN = True
    while FLAG_RUN:
        tcheck      = time.time()
        if True:
            file_age      = tcheck - tfile
            FLAG_TIME = (dtnews > 0)  and (file_age >= dtnews)
            FLAG_SIZE = (sizenewb > 0) and (bytes_written >= sizenewb)
            if(FLAG_TIME or FLAG_SIZE):
                f.close()
                data_stat = {'_deviceinfo': {}}
                data_stat['_deviceinfo']['filename'] = filename
                data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                data_stat['_deviceinfo']['closed'] = time.time()
                data_stat['_deviceinfo']['bytes_written'] = bytes_written
                data_stat['_deviceinfo']['packets_written'] = packets_written
                dataqueue.put(data_stat)
                [f,filename] = create_logfile(config,count)
                count += 1
                statistics = create_data_statistic_dict()
                tfile = tcheck   
                bytes_written         = 0
                packets_written       = 0
                data_stat = {'_deviceinfo': {}}
                data_stat['_deviceinfo']['filename'] = filename
                data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                data_stat['_deviceinfo']['created'] = time.time()
                data_stat['_deviceinfo']['bytes_written'] = bytes_written
                data_stat['_deviceinfo']['packets_written'] = packets_written
                dataqueue.put(data_stat)



        time.sleep(0.05)
        while(datainqueue.empty() == False):
            try:
                data = datainqueue.get(block=False)
                if (data is not None):
                    command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
                    #logger.debug('Got a command: {:s}'.format(str(data)))
                    if (command is not None):
                        if(command == 'stop'):
                            logger.debug(funcname + ' Stop command')
                            FLAG_RUN = False
                            f.close()
                            data_stat = {'_deviceinfo': {}}
                            data_stat['_deviceinfo']['filename'] = filename
                            data_stat['_deviceinfo']['filename_full'] = os.path.realpath(filename)
                            data_stat['_deviceinfo']['closed'] = time.time()
                            data_stat['_deviceinfo']['bytes_written'] = bytes_written
                            data_stat['_deviceinfo']['packets_written'] = packets_written
                            dataqueue.put(data_stat)
                            return

                statistics, status = do_data_statistics(data,statistics)
                yamlstr = yaml.dump(data,explicit_end=True,explicit_start=True)
                bytes_written         += len(yamlstr)
                packets_written       += 1
                bytes_written_total   += len(yamlstr)
                packets_written_total += 1
                f.write(yamlstr.encode('utf-8'))
                f.write(b'\0')
                if((time.time() - tflush) > config['dt_sync']):
                    f.flush()
                    os.fsync(f.fileno())
                    tflush = time.time()


                # Send statistics
                if ((time.time() - tupdate) > config['dt_update']):
                    tupdate = time.time()
                    data_stat = {'_deviceinfo':{}}
                    data_stat['_deviceinfo']['filename']        = filename
                    data_stat['_deviceinfo']['filename_full']   = os.path.realpath(filename)
                    data_stat['_deviceinfo']['bytes_written']   = bytes_written
                    data_stat['_deviceinfo']['packets_written'] = packets_written
                    dataqueue.put(data_stat)

            except Exception as e:
                logger.exception(e)
                logger.debug(funcname + ':Exception:' + str(e))
                #print(data)


#
#
# The init widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QGridLayout(self)
        self.device   = device
        self.redvypr  = device.redvypr
        self.label    = QtWidgets.QLabel("rawdatawriter setup")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet(''' font-size: 24px; font: bold''')
        self.config_widgets= [] # A list of all widgets that can only be used of the device is not started yet
        # Input output widget
        self.inlabel  = QtWidgets.QLabel("Input")
        self.inlist   = QtWidgets.QListWidget()
        self.adddeviceinbtn   = QtWidgets.QPushButton("Subscribe")
        self.adddeviceinbtn.clicked.connect(self.con_clicked)
        # The output widgets
        self.outlabel        = QtWidgets.QLabel("Logfile")
        self.outfilename     = QtWidgets.QLineEdit()
        # Checkboxes
        self.prefix_check    = QtWidgets.QCheckBox('Prefix')
        self.date_check      = QtWidgets.QCheckBox('Date/Time')
        self.count_check     = QtWidgets.QCheckBox('Counter')
        self.postfix_check   = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')

        try:
            filename = self.device.custom_config.filename
        except:
            filename = ''

        self.outfilename.setText(filename)
        
        self.folderbtn   = QtWidgets.QPushButton("Folder")
        self.config_widgets.append(self.folderbtn)
        self.folderbtn.clicked.connect(self.get_datafolder)
        
        # The rest
        #self.conbtn = QtWidgets.QPushButton("Connect logger to devices")
        #self.conbtn.clicked.connect(self.con_clicked)        
        self.startbtn = QtWidgets.QPushButton("Start logging")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred,QtWidgets.QSizePolicy.Expanding)


        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.dt_newfile = edit
        self.dt_newfile.setToolTip('Create a new file every N seconds.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.dt_newfile.setText(str(self.device.custom_config.dt_newfile))
        except Exception as e:
            self.dt_newfile.setText('0')
            
        # Delta t for new file
        edit = QtWidgets.QLineEdit(self)
        onlyInt = QtGui.QIntValidator()
        edit.setValidator(onlyInt)
        self.size_newfile = edit
        self.size_newfile.setToolTip('Create a new file every N bytes.\nFilename is "filenamebase"_yyyymmdd_HHMMSS_count."ext".\nUse 0 to disable feature.')
        try:
            self.size_newfile.setText(str(self.device.custom_config.size_newfile))
        except Exception as e:
            self.size_newfile.setText('0')
            
        self.newfiletimecombo = QtWidgets.QComboBox()
        self.newfiletimecombo.addItem('None')
        self.newfiletimecombo.addItem('seconds')
        self.newfiletimecombo.addItem('hours')
        self.newfiletimecombo.addItem('days')
        self.newfiletimecombo.setCurrentIndex(1)
            
        self.newfilesizecombo = QtWidgets.QComboBox()
        self.newfilesizecombo.addItem('None')
        self.newfilesizecombo.addItem('bytes')
        self.newfilesizecombo.addItem('kB')
        self.newfilesizecombo.addItem('MB')
        self.newfilesizecombo.setCurrentIndex(2)
        
        sizelabel = QtWidgets.QLabel('New file after')
         # File change layout
        self.newfilewidget = QtWidgets.QWidget()
        self.newfilelayout = QtWidgets.QFormLayout(self.newfilewidget)
        self.newfilelayout.addRow(sizelabel)
        self.newfilelayout.addRow(self.dt_newfile,self.newfiletimecombo)
        self.newfilelayout.addRow(self.size_newfile,self.newfilesizecombo)
        
        # Filenamelayout
        self.folder_text = QtWidgets.QLineEdit('')
        self.extension_text = QtWidgets.QLineEdit('redvypr_raw')
        self.prefix_text = QtWidgets.QLineEdit('')
        self.date_text = QtWidgets.QLineEdit('%Y-%m-%d_%H%M%S')
        self.count_text = QtWidgets.QLineEdit('04d')
        self.postfix_text = QtWidgets.QLineEdit('')
        self.gzip_text = QtWidgets.QLineEdit('gz')

        self.prefix_check = QtWidgets.QCheckBox('Prefix')
        self.date_check = QtWidgets.QCheckBox('Date/Time')
        self.count_check = QtWidgets.QCheckBox('Counter')
        self.postfix_check = QtWidgets.QCheckBox('Postfix')
        self.extension_check = QtWidgets.QCheckBox('Extension')
        self.gzip_check = QtWidgets.QCheckBox('gzip')
        # The outwidget
        self.outwidget = QtWidgets.QWidget()
        self.outlayout = QtWidgets.QGridLayout(self.outwidget)
        # Datafolder lineedit
        self.outlayout.addWidget(self.folderbtn, 0, 0)
        self.outlayout.addWidget(self.folder_text, 0, 1,1,3)
        # Checkboxes
        self.outlayout.addWidget(self.prefix_check, 1, 0)
        self.outlayout.addWidget(self.date_check, 1, 1)
        self.outlayout.addWidget(self.count_check, 1, 2)
        self.outlayout.addWidget(self.postfix_check, 1, 3)
        self.outlayout.addWidget(self.extension_check, 1, 4)
        self.outlayout.addWidget(self.gzip_check, 1, 5)

        self.outlayout.addWidget(self.prefix_text, 2, 0)
        self.outlayout.addWidget(self.date_text, 2, 1)
        self.outlayout.addWidget(self.count_text, 2, 2)
        self.outlayout.addWidget(self.postfix_text, 2, 3)
        self.outlayout.addWidget(self.extension_text, 2, 4)
        self.outlayout.addWidget(self.gzip_text, 2, 5)


        self.outlayout.addWidget(self.newfilewidget,4,0,1,4)

        #self.outlayout.addStretch(1)
            
        layout.addWidget(self.label,0,0,1,2)
        layout.addWidget(self.inlabel,1,0)         
        layout.addWidget(self.inlist,2,0)      
        layout.addWidget(self.adddeviceinbtn,3,0)      
        layout.addWidget(self.outlabel,1,1)
        layout.addWidget(self.outwidget,2,1,3,1)   
        layout.addWidget(self.startbtn,6,0,2,2)

        self.config_to_widgets()
        self.connect_widget_signals()
        # Connect the signals that notify a change of the connection
        self.device.redvypr.device_status_changed_signal.connect(self.update_device_list)
        #self.redvypr.devices_connected.connect
        self.device.subscription_changed_signal.connect(self.update_device_list)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

    def connect_widget_signals(self,connect=True):
        """
        Connects the signals of the widgets such that an update of the config is done

        Args:
            connect:

        Returns:

        """
        funcname = self.__class__.__name__ + '.connect_widget_signals():'
        logger.debug(funcname)
        if(connect):
            self.prefix_check.stateChanged.connect(self.update_device_config)
            self.postfix_check.stateChanged.connect(self.update_device_config)
            self.date_check.stateChanged.connect(self.update_device_config)
            self.count_check.stateChanged.connect(self.update_device_config)
            self.extension_check.stateChanged.connect(self.update_device_config)
            self.prefix_text.editingFinished.connect(self.update_device_config)
            self.postfix_text.editingFinished.connect(self.update_device_config)
            self.date_text.editingFinished.connect(self.update_device_config)
            self.count_text.editingFinished.connect(self.update_device_config)
            self.extension_text.editingFinished.connect(self.update_device_config)
            self.newfilesizecombo.currentIndexChanged.connect(self.update_device_config)
            self.newfiletimecombo.currentIndexChanged.connect(self.update_device_config)
            self.gzip_check.stateChanged.connect(self.update_device_config)
        else:
            self.gzip_check.stateChanged.disconnect()
            self.prefix_check.stateChanged.disconnect()
            self.postfix_check.stateChanged.disconnect()
            self.date_check.stateChanged.disconnect()
            self.count_check.stateChanged.disconnect()
            self.extension_check.stateChanged.disconnect()
            self.prefix_text.editingFinished.disconnect()
            self.postfix_text.editingFinished.disconnect()
            self.date_text.editingFinished.disconnect()
            self.count_text.editingFinished.disconnect()
            self.extension_text.editingFinished.disconnect()
            self.newfilesizecombo.currentIndexChanged.disconnect()
            self.newfiletimecombo.currentIndexChanged.disconnect()


    def get_datafolder(self):
        funcname = self.__class__.__name__ + '.get_datafolder():'
        logger.debug(funcname)
        retdata = QtWidgets.QFileDialog.getExistingDirectory(self,"Choose datafolder")
        #print('Datafolder',retdata)
        datafolder = retdata
        if datafolder:
            self.folder_text.setText(datafolder)
            
    def con_clicked(self):
        funcname = self.__class__.__name__ + '.con_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if(button == self.adddeviceinbtn):
            self.connect.emit(self.device) # The connect signal is connected with connect_device that will open a subscribe/connect widget
            self.update_device_list()
                            
    def update_device_list(self):
        funcname = self.__class__.__name__ + '.update_device_list():'
        logger.debug(funcname)
        #print('Devices',devicestr_provider,devicestr_receiver)
        raddresses = self.device.get_subscribed_deviceaddresses()
        #print('Deviceaddresses',raddresses)
        self.inlist.clear()
        for raddr in raddresses:
            self.inlist.addItem(raddr.address_str)


    def config_to_widgets(self):
        """
        Updates the widgets according to the device config

        Returns:

        """
        funcname = self.__class__.__name__ + '.config_to_widgets():'
        logger.debug(funcname)

        config = self.device.custom_config
        #print('config',config)
        self.dt_newfile.setText(str(config.dt_newfile))
        for i in range(self.newfiletimecombo.count()):
            self.newfiletimecombo.setCurrentIndex(i)
            if(self.newfiletimecombo.currentText().lower() == config.dt_newfile_unit):
                break

        for i in range(self.newfilesizecombo.count()):
            self.newfilesizecombo.setCurrentIndex(i)
            if (self.newfilesizecombo.currentText().lower() == config.size_newfile_unit):
                break

        self.size_newfile.setText(str(config.size_newfile))

        if len(config.datafolder)>0:
            self.folder_text.setText(config.datafolder)
        # Update filename and checkboxes
        filename_all = []
        filename_all.append([config.fileextension,self.extension_text,self.extension_check])
        filename_all.append([config.fileprefix,self.prefix_text,self.prefix_check])
        filename_all.append([config.filepostfix,self.postfix_text,self.postfix_check])
        filename_all.append([config.filedateformat,self.date_text,self.date_check])
        filename_all.append([config.filecountformat,self.count_text,self.count_check])
        filename_all.append([config.filegzipformat, self.gzip_text, self.gzip_check])
        for i in range(len(filename_all)):
            widgets = filename_all[i]
            if(len(widgets[0])==0):
                widgets[2].setChecked(False)
                widgets[1].setText('')
            else:
                widgets[2].setChecked(True)
                widgets[1].setText(widgets[0])

    def widgets_to_config(self,config):
        """
        Reads the widgets and creates a config
        Returns:
            config: Config dictionary
        """
        funcname = self.__class__.__name__ + '.widgets_to_config():'
        logger.debug(funcname)
        config.datafolder = self.folder_text.text()
        config.dt_newfile = int(self.dt_newfile.text())
        config.dt_newfile_unit = self.newfiletimecombo.currentText()
        config.size_newfile = int(self.size_newfile.text())
        config.size_newfile_unit = self.newfilesizecombo.currentText()

        if(self.extension_check.isChecked()):
            config.fileextension = self.extension_text.text()
        else:
            config.fileextension = ''

        if(self.prefix_check.isChecked()):
            config.fileprefix = self.prefix_text.text()
        else:
            config.fileprefix = ''

        if(self.postfix_check.isChecked()):
            config.filepostfix = self.postfix_text.text()
        else:
            config.filepostfix = ''

        if(self.date_check.isChecked()):
            config.filedateformat    = self.date_text.text()
        else:
            config.filedateformat = ''

        if(self.count_check.isChecked()):
            config.filecountformat = self.count_text.text()
        else:
            config.filecountformat = ''

        if (self.gzip_check.isChecked()):
            config.filegzipformat = self.gzip_text.text()
        else:
            config.filegzipformat = ''


        #print('Config',config)
        return config

    def update_device_config(self):
        """
        Updates the device config based on the widgets
        Returns:

        """
        funcname = self.__class__.__name__ + '.update_device_config():'
        logger.debug(funcname)
        self.widgets_to_config(self.device.custom_config)

    def start_clicked(self):
        funcname = self.__class__.__name__ + '.start_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if button.isChecked():
            logger.debug(funcname + "button pressed")
            self.update_device_config()
            logger.debug(funcname + " Clearing datainqueue")
            q = self.device.datainqueue
            while not q.empty():
                try:
                    q.get(block=False)
                except queue.Empty:
                    continue
                q.task_done()
            self.device.thread_start()
        else:
            logger.debug(funcname + 'button released')
            self.device.thread_stop()

            
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']

            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                for w in self.config_widgets:
                    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                for w in self.config_widgets:
                    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self):
        super(QtWidgets.QWidget, self).__init__()
        layout          = QtWidgets.QVBoxLayout(self)
        hlayout         = QtWidgets.QHBoxLayout()
        self.filetable  = QtWidgets.QTableWidget()
        headers = ['Filename','Bytes written','Packets written','Date created','Date closed','Full filepath']
        self.filetable.setColumnCount(len(headers))
        self.filetable.setHorizontalHeaderLabels(headers)
        self.filetable.resizeColumnsToContents()
        self.filelab= QtWidgets.QLabel("File: ")
        self.byteslab   = QtWidgets.QLabel("Bytes written: ")
        self.packetslab = QtWidgets.QLabel("Packets written: ")
        hlayout.addWidget(self.byteslab)
        hlayout.addWidget(self.packetslab)
        layout.addWidget(self.filelab)        
        layout.addLayout(hlayout)
        layout.addWidget(self.filetable)
        #self.text.insertPlainText("hallo!")        

    def update_data(self,data):
        funcname = __name__ + '.update()'
        #print('data',data)
        try:
            filename_table = self.filetable.item(0,0).text()
        except:
            filename_table = ''

        #print('Filename table',filename_table,data['_deviceinfo']['filename'])
        try:
            tclose = datetime.datetime.fromtimestamp(data['_deviceinfo']['closed']).strftime(
                '%d-%m-%Y %H:%M:%S')
            item = QtWidgets.QTableWidgetItem(tclose)
            self.filetable.setItem(0, 4, item)
        except:
            pass

        if filename_table != data['_deviceinfo']['filename']:
            self.filetable.insertRow(0)
            try:
                #headers = ['Filename', 'Date created', 'Bytes written', 'Date closed']
                item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename'])
                self.filetable.setItem(0,0,item)
                item = QtWidgets.QTableWidgetItem(data['_deviceinfo']['filename_full'])
                self.filetable.setItem(0, 5, item)
                tcreate = datetime.datetime.fromtimestamp(data['_deviceinfo']['created']).strftime('%d-%m-%Y %H:%M:%S')
                item = QtWidgets.QTableWidgetItem(tcreate)
                self.filetable.setItem(0, 3, item)
            except Exception as e:
                logger.exception(e)

        item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['bytes_written']))
        self.filetable.setItem(0, 1, item)
        item = QtWidgets.QTableWidgetItem(str(data['_deviceinfo']['packets_written']))
        self.filetable.setItem(0, 2, item)



        self.filetable.resizeColumnsToContents()


        self.filelab.setText("File: {:s}".format(data['_deviceinfo']['filename']))
        self.byteslab.setText("Bytes written: {:d}".format(data['_deviceinfo']['bytes_written']))
        self.packetslab.setText("Packets written: {:d}".format(data['_deviceinfo']['packets_written']))


        #self.text.insertPlainText(str(data['_deviceinfo']['data']))
        



import redvypr
import time
redvypr.logger.setLevel('DEBUG')

r = redvypr.Redvypr(hostname='metadatatest',loglevel='DEBUG')
devicemodulename = r.get_devicemodulename_from_str('test_device')
dev = r.add_device(devicemodulename=devicemodulename)

print('Device',dev)
dev.thread_start()
time.sleep(2)
dev.thread_stop()
time.sleep(0.5)

print('----------------------')
print('Device metadata')
print(dev.statistics['metadata'])
print('Device metadata done')
print('----------------------')

raddrs = []
raddrs.append(redvypr.RedvyprAddress('*'))
raddrs.append(redvypr.RedvyprAddress('/k:["data_list_list"]'))
raddrs.append(redvypr.RedvyprAddress("/k:['data_list_list'][0]"))
raddrs.append(redvypr.RedvyprAddress("/k:['data_list_list'][1]"))
raddrs.append(redvypr.RedvyprAddress('/k:["data_list_poly"][0]'))
raddrs.append(redvypr.RedvyprAddress('/k:["data_dict_list"]["temp"]'))
raddrs.append(redvypr.RedvyprAddress("/k:['data_dict_list']['temp'][0]"))
raddrs.append(redvypr.RedvyprAddress('/k:["data_dict_list"]["pressure"][0]'))

#print('Raddr1',raddr1.datakeyeval)
for raddr in raddrs:
    meta = dev.get_metadata(raddr)
    print('-----------------------')
    print('Metadata for address {}: {}'.format(raddr,meta))

